with Rounding, Rounding.FPU, Globals, Strings;
use Globals, Strings;

pragma Elaborate_All (Rounding.FPU,Globals, Strings);

package body TCosSins1 is

  procedure Show_C1_Rounding is
    use Rounding,Rounding.FPU;
    M: constant Rounding_Mode := Get_Rounding_Mode;
    Prec: constant Positive := Get_Precision(SOne/Three);
  begin
    Show1("C1:      Precision = ",Prec);        --- set precision before elaboration
    Show0("C1:  Rounding_Mode =  " & Strng(M));
  end Show_C1_Rounding;

  procedure Show_C1_Param(All_Param: in Boolean := True) is
    SInfo: constant Scalar_Info := Info(SZero);
  begin
    Show1("C1:           FDeg = ",FDeg);
    Show1("C1:         Radius = ",Rad);
    Show1("C1:          Trunc =  ",Trunc);
    Show1("C1: Default_Parity = ",Default_Parity);
    if All_Param then
      Show0("C1:        RepType =  " & SInfo.RepType);
      Show1("C1:         NTasks = ",NTasks);
      Show_C1_Rounding;
    end if;
  end Show_C1_Param;

  function CosSin(P: Parity) return String is
  begin
    if P=0 then return "Cos"; end if;
    return "Sin";
  end CosSin;

  function Is_Zero(P: Parity; C: FPoly1) return Boolean is
  begin
    if FDeg=0 then return IsZero(C(0)); end if;
    if P=0 and then not IsZero(C(0)) then return False; end if;
    for I in 1 .. FDeg loop
      if not IsZero(C(I)) then return False; end if;
    end loop;
    return True;
  end Is_Zero;

  function Is_Zero(E: Errs1) return Boolean is
  begin
    if Errs then
      for I in EFreq loop
        if E(I) /= Zero then return False; end if;
      end loop;
    end if;
    return True;
  end Is_Zero;

  procedure LiftE0(E: in out Errs1) is
    --- use only if Not_Trunc and FDeg>0
  begin
    if (E(0) /= Zero) then
      E(1) := E(1)+E(0);
      E(0) := Zero;
    end if;
  end LiftE0;

  procedure XPowers(X: in Scalar; V: in out SVector) is
    --- V(N)=X^N
  begin
    Copy(SOne,V(0));
    Copy(X,V(1));
    for N in 2 .. V'Last loop
      Prod(V(N/2),V(N-N/2),V(N));
    end loop;
  end XPowers;

  procedure CSPowers(X: in Scalar; C,S: in out SVector) is
    --- C(I)=Cos(I*X), S(I)=Sin(I*X)
    --- assuming 0 .. Last
    Last: constant Positive := IMin(C'Last,S'Last);
    Tmp1: Scalar renames C(0);
    Tmp2: Scalar renames S(0);
    I,J: Integer;
  begin
    Cos(X,C(1));
    Sin(X,S(1));
    for N in 2 .. Last loop
      I := N/2;
      J := N-I;
      Prod(C(I),C(J),Tmp1);
      Prod(S(I),S(J),Tmp2);
      Diff(Tmp1,Tmp2,C(N)); -- Cn = Ci*Cj-Si*Sj
      Prod(C(I),S(J),Tmp1);
      Prod(S(I),C(J),Tmp2);
      Sum(Tmp1,Tmp2,S(N));  -- Sn = Ci*Sj+Si*Cj
    end loop;
    Copy(SOne,C(0));
    SetZero(S(0));
  end CSPowers;

  ----------------------------------

  function IsZero(P: CosSin1) return Boolean is
  begin
    if FDeg=0 then return IsZero(P.C(0)); end if;
    declare
      C: FPoly1 renames P.C;
    begin
      for I in P.P .. FDeg loop
        if not IsZero(C(I)) then return False; end if;
      end loop;
    end;
    if Not_Trunc then
      declare
        E: Errs1 renames P.E;
      begin
        for I in EFreq loop
          if E(I) /= Zero then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end IsZero;

  procedure Parities(P: in CosSin1; Even,Odd: out Boolean) is
    --- Even=Odd=True means P=0
    --- Even=Odd=False cannot occur
    C: FPoly1 renames P.C;
  begin
    Even := (P.P=0);
    Odd := not Even;
    for I in P.P .. FDeg loop
      if not IsZero(C(I)) then return; end if;
    end loop;
    if Errs then
      declare
        E: Errs1 renames P.E;
      begin
        for I in EFreq loop
          if E(I) /= Zero then return; end if;
        end loop;
      end;
    end if;       -- P=0
    Even := True;
    Odd := True;
  end Parities;

  function IsSharp(P: CosSin1) return Boolean is
  begin
    if Trunc then return True; end if;
    if FDeg=0 then return IsSharp(P.C(0)); end if;
    declare
      C: FPoly1 renames P.C;
    begin
      for I in P.P .. FDeg loop
        if not IsSharp(C(I)) then return False; end if;
      end loop;
    end;
    declare
      E: Errs1 renames P.E;
    begin
      for I in EFreq loop
        if E(I) /= Zero then return False; end if;
      end loop;
    end;
    return True;
  end IsSharp;

  function IsPoly(P: CosSin1) return Boolean is
  begin
    if Errs then
      declare
        E: Errs1 renames P.E;
      begin
        for I in EFreq loop
          if E(I) /= Zero then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end IsPoly;

  function "="(P1,P2: CosSin1) return Boolean is
  begin
    if P1.P /= P2.P then return False; end if;
    if FDeg=0 then return P1.C(0)=P2.C(0); end if;
    declare
      C1: FPoly1 renames P1.C;
      C2: FPoly1 renames P2.C;
    begin
      for I in P1.P .. FDeg loop
        if C1(I) /= C2(I) then return False; end if;
      end loop;
    end;
    if Not_Trunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
      begin
        for I in EFreq loop
          if E1(I)/=E2(I) then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end "=";

  procedure FillZero(P: in out CosSin1) is
  begin
    if FDeg=0 then
      SetZero(P.C(0));
    else
      declare
        C: FPoly1 renames P.C;
      begin
        for I in FFreq loop SetZero(C(I)); end loop;
      end;
      if Not_Trunc then P.E := (others => Zero); end if;
    end if;
  end FillZero;

  procedure SetZero(Par: in Parity; P: in out CosSin1) is
  begin
    if FDeg=0 then
      P.P := 0;
      SetZero(P.C(0));
    else
      P.P := Par;
      declare
        C: FPoly1 renames P.C;
      begin
        for I in FFreq loop SetZero(C(I)); end loop;
      end;
      if Not_Trunc then P.E := (others => Zero); end if;
    end if;
  end SetZero;

  procedure SetZero(P: in out CosSin1) is
  begin
    if FDeg=0 then
      P.P := 0;
      SetZero(P.C(0));
    else
      P.P := Default_Parity;
      declare
        C: FPoly1 renames P.C;
      begin
        for I in FFreq loop SetZero(C(I)); end loop;
      end;
      if Not_Trunc then P.E := (others => Zero); end if;
    end if;
  end SetZero;

  procedure Copy(P1: in CosSin1; P2: in out CosSin1) is
  begin
    if FDeg=0 then
      P2.P := 0;
      Copy(P1.C(0),P2.C(0));
    else
      P2.P := P1.P;
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
      begin
        for I in P1.P .. FDeg loop Copy(C1(I),C2(I)); end loop;
      end;
      if Not_Trunc then P2.E := P1.E; end if;
    end if;
  end Copy;

  procedure Swap(P1,P2: in out CosSin1) is
  begin
    if FDeg=0 then
      Swap(P1.C(0),P2.C(0));
    else
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
      begin
        if P1.P=1 then SetZero(C1(0)); end if;
        if P2.P=1 then SetZero(C2(0)); end if;
        if P1.P/=P2.P then
          P1.P := 1-P1.P;
          P2.P := 1-P2.P;
        end if;
        for I in FFreq loop Swap(C1(I),C2(I)); end loop;
      end;
      if Not_Trunc then
        declare
          E1: constant Errs1 := P1.E;
        begin
          P1.E := P2.E;
          P2.E := E1;
        end;
      end if;
    end if;
  end Swap;

  ----------------------------------

  function Center0(P: CosSin1) return Boolean is
    --- symmetric about zero
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop
      if not Center0(C(I)) then return False; end if;
    end loop;
    return True;
  end Center0;

  function Contains0(P: CosSin1) return Logical is
    --- very crude
  begin
    if FDeg=0 then return Contains0(P.C(0)); end if;
    declare
      C: FPoly1 renames P.C;
      L: Logical := True;
      LI: Logical;
    begin
      for I in P.P .. FDeg loop
        LI := Contains0(C(I));
        if LI=False then return False; end if;
        if LI=Uncertain then L := Uncertain; end if;
      end loop;
      return L;
    end;
  end Contains0;

  procedure BallAt0(R: in Flt; Par: in Parity; P: in out CosSin1) is
  begin
    if FDeg=0 then
      P.P := 0;
      BallAt0(R,P.C(0));
    else
      SetZero(Par,P);
      if Not_Trunc then
        P.E(0) := R;
        if P.P=1 then LiftE0(P.E); end if;
      end if;
    end if;
  end BallAt0;

  procedure BallAt0(R: in Flt; P: in out CosSin1) is
  begin
    if FDeg=0 then
      BallAt0(R,P.C(0));
    else
      SetZero(Default_Parity,P);
      if Not_Trunc then
        P.E(0) := R;
        if P.P=1 then LiftE0(P.E); end if;
      end if;
    end if;
  end BallAt0;

  function BallAt0(R: Flt) return CosSin1 is
    P: CosSin1;
  begin
    BallAt0(R,P);
    return P;
  end BallAt0;

  procedure ToErr(P: in out CosSin1) is
    --- zero-centered set containing P
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop ToErr(C(I)); end loop;
  end ToErr;

  procedure ToErr(P1: in CosSin1; P2: in out CosSin1) is
    --- zero-centered set containing P1
  begin
    if FDeg=0 then
      P2.P := 0;
      ToErr(P1.C(0),P2.C(0));
    else
      P2.P := P1.P;
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
      begin
        if P1.P=0 then SetZero(C2(0)); else ToErr(C1(0),C2(0)); end if;
        for I in 1 .. FDeg loop ToErr(C1(I),C2(I)); end loop;
      end;
      if Not_Trunc then P2.E := P1.E; end if;
    end if;
  end ToErr;

  function ToErr(P: CosSin1) return CosSin1 is
    --- zero-centered set containing P
    Q: CosSin1;
  begin
    ToErr(P,Q);
    return Q;
  end ToErr;

  procedure Center(P: in out CosSin1) is
  begin
    if Not_Trunc then
      if FDeg=0 then
        Center(P.C(0));
      else
        declare
          C: FPoly1 renames P.C;
        begin
          for I in P.P .. FDeg loop Center(C(I)); end loop;
        end;
        P.E := (others => Zero);
      end if;
    end if;
  end Center;

  procedure Center(P1: in CosSin1; P2: in out CosSin1) is
  begin
    if FDeg=0 then
      P2.P := 0;
      Center(P1.C(0),P2.C(0));
    else
      P2.P := P1.P;
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
      begin
        if P1.P=0 then
          Center(C1(0),C2(0));
        else
          SetZero(C2(0));
        end if;
        for I in 1 .. FDeg loop Center(C1(I),C2(I)); end loop;
      end;
      if Not_Trunc then P2.E := (others => Zero); end if;
    end if;
  end Center;

  function Center(P: CosSin1) return CosSin1 is
    Q: CosSin1;
  begin
    Center(P,Q);
    return Q;
  end Center;

  procedure ModCenter(P: in out CosSin1) is
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop ModCenter(C(I)); end loop;
  end ModCenter;

  procedure ModCenter(P1: in CosSin1; P2: in out CosSin1) is
  begin
    if FDeg=0 then
      P2.P := 0;
      ModCenter(P1.C(0),P2.C(0));
    else
      P2.P := P1.P;
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
      begin
        for I in P1.P .. FDeg loop ModCenter(C1(I),C2(I)); end loop;
      end;
      if Not_Trunc then P2.E := P1.E; end if;
    end if;
  end ModCenter;

  function ModCenter(P: CosSin1) return CosSin1 is
    Q: CosSin1;
  begin
    ModCenter(P,Q);
    return Q;
  end ModCenter;

  procedure SplitErr(P: in CosSin1; PC,PE: in out CosSin1) is
    --- split P into a sharp PC and an error PE
  begin
    Center(P,PC);
    ModCenter(P,PE);
  end SplitErr;

  procedure ErrMult(R: in Radius; P: in out CosSin1) is
  begin
    if Not_Trunc then
      if FDeg=0 then
        ErrMult(R,P.C(0));
      else
        declare
          C: FPoly1 renames P.C;
          E: Errs1 renames P.E;
        begin
          for I in P.P .. FDeg loop ErrMult(R,C(I)); end loop;
          for I in EFreq loop E(I) := R*E(I); end loop;
        end;
      end if;
    end if;
  end ErrMult;

  ------------------------------------------------------------

  procedure Neg(P: in out CosSin1) is
  begin
    if FDeg=0 then
      Neg(P.C(0));
    else
      declare
        C: FPoly1 renames P.C;
      begin
        for I in P.P .. FDeg loop Neg(C(I)); end loop;
      end;
    end if;
  end Neg;

  procedure Neg(P1: in CosSin1; P2: in out CosSin1) is
  begin
    if FDeg=0 then
      P2.P := 0;
      Neg(P1.C(0),P2.C(0));
    else
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
      begin
        if P1.P=0 then
          P2.P := 0;
          Neg(C1(0),C2(0));
        else
          P2.P := 1;
          SetZero(C2(0));
        end if;
        for I in 1 .. FDeg loop Neg(C1(I),C2(I)); end loop;
      end;
      if Not_Trunc then P2.E := P1.E; end if;
    end if;
  end Neg;

  function "-"(P: CosSin1) return CosSin1 is
    Q: CosSin1;
  begin
    Neg(P,Q);
    return Q;
  end "-";

  procedure Add(I: in Integer; P: in out CosSin1) is
  begin
    if P.P=0 then
      Add(I,P.C(0));
    elsif I=0 then
      null;
    elsif IsZero(P) then
      P.P := 0;
      Add(I,P.C(0));
    else
      raise Parity_Error;
    end if;
  end Add;

  procedure Add(S: in Scalar; P: in out CosSin1) is
  begin
    if P.P=0 then
      Add(S,P.C(0));
    elsif IsZero(S) then
      null;
    elsif IsZero(P) then
      P.P := 0;
      Add(S,P.C(0));
    else
      raise Parity_Error;
    end if;
  end Add;

  procedure Add(P1: in CosSin1; P2: in out CosSin1) is
  begin
    if FDeg=0 then
      Add(P1.C(0),P2.C(0));
    elsif P1.P=P2.P then
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
      begin
        for I in P1.P .. FDeg loop Add(C1(I),C2(I)); end loop;
      end;
      if Not_Trunc then
        declare
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
        begin
          for I in EFreq loop E2(I) := E2(I)+E1(I); end loop;
        end;
      end if;
    elsif IsZero(P2) then
      Copy(P1,P2);
    elsif not IsZero(P1) then
      raise Parity_Error;
    end if;
  end Add;

  procedure Sum(P1,P2: in CosSin1; P3: in out CosSin1) is
  begin
    if FDeg=0 then
      P3.P := 0;
      Sum(P1.C(0),P2.C(0),P3.C(0));
    elsif P1.P=P2.P then
      P3.P := P1.P;
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
        C3: FPoly1 renames P3.C;
      begin
        for I in P1.P .. FDeg loop Sum(C1(I),C2(I),C3(I)); end loop;
      end;
      if Not_Trunc then
        declare
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
          E3: Errs1 renames P3.E;
        begin
          for I in EFreq loop E3(I) := E1(I)+E2(I); end loop;
        end;
      end if;
    elsif IsZero(P1) then
      Copy(P2,P3);
    elsif IsZero(P2) then
      Copy(P1,P3);
    else
      raise Parity_Error;
    end if;
  end Sum;

  function "+"(P1,P2: CosSin1) return CosSin1 is
    P3: CosSin1;
  begin
    Sum(P1,P2,P3);
    return P3;
  end "+";

  procedure Sub(P1: in CosSin1; P2: in out CosSin1) is
  begin
    if FDeg=0 then
      Sub(P1.C(0),P2.C(0));
    elsif P1.P=P2.P then
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
      begin
        for I in P1.P .. FDeg loop Sub(C1(I),C2(I)); end loop;
      end;
      if Not_Trunc then
        declare
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
        begin
          for I in EFreq loop E2(I) := E2(I)+E1(I); end loop;
        end;
      end if;
    elsif IsZero(P2) then
      Neg(P1,P2);
    elsif not IsZero(P1) then
      raise Parity_Error;
    end if;
  end Sub;

  procedure Diff(P1,P2: in CosSin1; P3: in out CosSin1) is
  begin
    if FDeg=0 then
      P3.P := 0;
      Diff(P1.C(0),P2.C(0),P3.C(0));
    elsif P1.P=P2.P then
      P3.P := P1.P;
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
        C3: FPoly1 renames P3.C;
      begin
        for I in P1.P .. FDeg loop Diff(C1(I),C2(I),C3(I)); end loop;
      end;
      if Not_Trunc then
        declare
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
          E3: Errs1 renames P3.E;
        begin
          for I in EFreq loop E3(I) := E1(I)+E2(I); end loop;
        end;
      end if;
    elsif IsZero(P2) then
      Copy(P1,P3);
    elsif IsZero(P1) then
      Neg(P2,P3);
    else
      raise Parity_Error;
    end if;
  end Diff;

  function "-"(P1,P2: CosSin1) return CosSin1 is
    P3: CosSin1;
  begin
    Diff(P1,P2,P3);
    return P3;
  end "-";

  procedure Mult(R: in Flt; P: in out CosSin1) is
  begin
    if FDeg=0 then
      Mult(R,P.C(0));
    elsif R=Zero then
      FillZero(P);
    else
      declare
        C: FPoly1 renames P.C;
        Tmp: Scalar;
      begin
        for I in P.P .. FDeg loop Mult(R,C(I),Tmp); end loop;
      end;
      if Not_Trunc then
        declare
          E: Errs1 renames P.E;
          A: constant Flt := abs(R);
        begin
          for I in EFreq loop E(I) := A*E(I); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Prod(R: in Flt; P1: in CosSin1; P2: in out CosSin1) is
  begin
    if FDeg=0 then
      P2.P := 0;
      Prod(R,P1.C(0),P2.C(0));
    elsif R=One then
      Copy(P1,P2);
    elsif R=NegOne then
      Neg(P1,P2);
    else
      P2.P := P1.P;
      if R=Zero then
        FillZero(P2);
      else
        declare
          C1: FPoly1 renames P1.C;
          C2: FPoly1 renames P2.C;
        begin
          for I in P1.P .. FDeg loop Prod(R,C1(I),C2(I)); end loop;
        end;
        if Not_Trunc then
          declare
            E1: Errs1 renames P1.E;
            E2: Errs1 renames P2.E;
            A: constant Flt := abs(R);
          begin
            for I in EFreq loop E2(I) := A*E1(I); end loop;
          end;
        end if;
      end if;
    end if;
  end Prod;

  function "*"(R: Flt; P: CosSin1) return CosSin1 is
    Q: CosSin1;
  begin
    Prod(R,P,Q);
    return Q;
  end "*";

  procedure AddProd(R: in Flt; P1: in CosSin1; P2: in out CosSin1) is
  begin
    if R=Zero then
      null;
    elsif FDeg=0 then
      AddProd(R,P1.C(0),P2.C(0));
    elsif R=One then
      Add(P1,P2);
    elsif R=NegOne then
      Sub(P1,P2);
    elsif P1.P=P2.P then
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
        Tmp: Scalar;
      begin
        for I in P1.P .. FDeg loop AddProd(R,C1(I),C2(I),Tmp); end loop;
      end;
      if Not_Trunc then
        declare
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
          A: constant Flt := abs(R);
        begin
          for I in EFreq loop E2(I) := E2(I)+A*E1(I); end loop;
        end;
      end if;
    elsif IsZero(P2) then
      Prod(R,P1,P2);
    elsif not IsZero(P1) then
      raise Parity_Error;
    end if;
  end AddProd;

  procedure Mult(Q: in Rational; P: in out CosSin1) is
  begin
    if FDeg=0 then
      Mult(Q,P.C(0));
    elsif LNum(Q)=0 then
      FillZero(P);
    else
      declare
        C: FPoly1 renames P.C;
      begin
        for I in P.P .. FDeg loop Mult(Q,C(I)); end loop;
      end;
      if not IsPoly(P) then
        declare
          A: constant Flt := Flt(abs(LNum(Q)))/Flt(LDen(Q));
          E: Errs1 renames P.E;
        begin
          for I in EFreq loop E(I) := A*E(I); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Mult(S: in Scalar; P: in out CosSin1) is
  begin
    if FDeg=0 then
      Mult(S,P.C(0));
    elsif IsZero(S) then
      FillZero(P);
    else
      declare
        C: FPoly1 renames P.C;
      begin
        for I in P.P .. FDeg loop Mult(S,C(I)); end loop;
      end;
      if Not_Trunc then
        declare
          A: constant Flt := MaxNorm(S);
          E: Errs1 renames P.E;
        begin
          for I in EFreq loop E(I) := A*E(I); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Prod(S: in Scalar; P1: in CosSin1; P2: in out CosSin1) is
  begin
    if FDeg=0 then
      P2.P := 0;
      Prod(S,P1.C(0),P2.C(0));
    else
      P2.P := P1.P;
      if IsZero(S) then
        FillZero(P2);
      else
        declare
          C1: FPoly1 renames P1.C;
          C2: FPoly1 renames P2.C;
        begin
          for I in P1.P .. FDeg loop Prod(S,C1(I),C2(I)); end loop;
        end;
        if Not_Trunc then
          declare
            A: constant Flt := MaxNorm(S);
            E1: Errs1 renames P1.E;
            E2: Errs1 renames P2.E;
          begin
            for I in EFreq loop E2(I) := A*E1(I); end loop;
          end;
        end if;
      end if;
    end if;
  end Prod;

  function "*"(S: Scalar; P: CosSin1) return CosSin1 is
    Q: CosSin1;
  begin
    Prod(S,P,Q);
    return Q;
  end "*";

  procedure AddProd(S: in Scalar; P1: in CosSin1; P2: in out CosSin1) is
  begin
    if FDeg=0 then
      AddProd(S,P1.C(0),P2.C(0));
    elsif IsZero(S) then
      null;
    elsif P1.P=P2.P then
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
        Tmp: Scalar;
      begin
        for I in P1.P .. FDeg loop
          if not IsZero(C1(I)) then AddProd(S,C1(I),C2(I),Tmp); end if;
        end loop;
      end;
      if Not_Trunc then
        declare
          A: constant Flt := MaxNorm(S);
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
        begin
          for I in EFreq loop E2(I) := E2(I)+A*E1(I); end loop;
        end;
      end if;
    elsif IsZero(P2) then
      Prod(S,P1,P2);
    elsif not IsZero(P1) then
      raise Parity_Error;
    end if;
  end AddProd;

  procedure Mult(P1: in CosSin1; P2,Tmp: in out CosSin1) is
  begin
    Prod(P1,P2,Tmp);
    Copy(Tmp,P2);
  end Mult;

  procedure Mult(P1: in CosSin1; P2: in out CosSin1) is
  begin
    if FDeg=0 then
      Mult(P1.C(0),P2.C(0));
    else
      declare
        Tmp: LP_Pointer;
      begin
        PPool.Allocate(Tmp);
        Prod(P1,P2,Tmp.all.Data);
        Copy(Tmp.all.Data,P2);
        PPool.Recycle(Tmp);
      end;
    end if;
  end Mult;

  --  2*cos(m)*cos(n) =  cos(m-n) + cos(m+n)
  --  2*sin(m)*sin(n) =  cos(m-n) - cos(m+n)
  --  2*cos(m)*sin(n) = -sin(m-n) + sin(m+n)
  --  2*sin(m)*cos(n) =  sin(m-n) + sin(m+n)

  procedure CCProd2(P1P,P2P: in Parity; C1,C2: in FPoly1; A1,A2: in CNorms; P3: in out CosSin1) is
    --- add twice C1*C2
    --- does not set P3.P or P3.R
    --- for efficiency, check first whether C1=0 or C2=0
    C3: FPoly1 renames P3.C;
    E3: Errs1 renames P3.E;
    K: Integer;
    Tmp: Scalar;
  begin
    if P1P=0 then
      if P2P=0 then                           --- 2*cos(m)*cos(n) = cos(m-n) + cos(m+n)
        for M in 0 .. FDeg loop
          if A1(M) /= Zero then
            for N in 0 .. FDeg loop
              if A2(N) /= Zero then
                K := M-N;                               --- freq diff
                AddProd(C1(M),C2(N),C3(abs(K)),Tmp);
                K := M+N;                               --- freq sum
                if K <= FDeg then
                  AddProd(C1(M),C2(N),C3(K),Tmp);
                elsif Not_Trunc then
                  E3(K) := E3(K)+A1(M)*A2(N); --- norms
                end if;
              end if;
            end loop;
          end if;
        end loop;
      else                                 --- 2*cos(m)*sin(n) = -sin(m-n) + sin(m+n)
        for M in 0 .. FDeg loop
          if A1(M) /= Zero then
            for N in 1 .. FDeg loop
              if A2(N) /= Zero then
                K := M-N;                            --- freq diff
                if K>0 then
                  SubProd(C1(M),C2(N),C3(K),Tmp);    ---    M>N
                elsif K<0 then
                  AddProd(C1(M),C2(N),C3(-K),Tmp);   ---    M<N
                end if;
                K := M+N;                            --- freq sum
                if K <= FDeg then
                  AddProd(C1(M),C2(N),C3(K),Tmp);
                elsif Not_Trunc then
                  E3(K) := E3(K)+A1(M)*A2(N); --- norms
                end if;
              end if;
            end loop;
          end if;
        end loop;
      end if;
    elsif P2P=0 then                    --- 2*sin(m)*cos(n) = sin(m-n) + sin(m+n)
      for M in 1 .. FDeg loop
        if A1(M) /= Zero then
          for N in 0 .. FDeg loop
            if A2(N) /= Zero then
              K := M-N;                            --- freq diff
              if K>0 then
                AddProd(C1(M),C2(N),C3(K),Tmp);    ---    M>N
              elsif K<0 then
                SubProd(C1(M),C2(N),C3(-K),Tmp);   ---    M<N
              end if;
              K := M+N;                            --- freq sum
              if K <= FDeg then
                AddProd(C1(M),C2(N),C3(K),Tmp);
              elsif Not_Trunc then
                E3(K) := E3(K)+A1(M)*A2(N); --- norms
              end if;
            end if;
          end loop;
        end if;
      end loop;
    else                                 --- 2*sin(m)*sin(n) =  cos(m-n) - cos(m+n)
      for M in 1 .. FDeg loop
        if A1(M) /= Zero then
          for N in 1 .. FDeg loop
            if A2(N) /= Zero then
              K := M-N;                             --- freq diff
              AddProd(C1(M),C2(N),C3(abs(K)),Tmp);
              K := M+N;                             --- freq sum
              if K <= FDeg then
                SubProd(C1(M),C2(N),C3(K),Tmp);
              elsif Not_Trunc then
                E3(K) := E3(K)+A1(M)*A2(N); --- norms
              end if;
            end if;
          end loop;
        end if;
      end loop;
    end if;
  end CCProd2;

  procedure CEProd2(P1P: in Parity; A1: in CNorms; E2: in Errs1; E3: in out Errs1) is
    --- add twice A1*E2
    --- for efficiency, check first whether A1=0 or E2=0
    E3F: Radius renames E3(0);
    E3L: Radius renames E3(EDeg);
    K: Integer;
  begin
    for M in P1P .. FDeg loop
      if A1(M) /= Zero then
        for N in EFreq loop
          if E2(N) /= Zero then
            K := N-M;                                      --- freq diff
            if K<0 then                                    ---   M>N
              E3F := E3F+InvRadPower(2*N)*A1(M)*E2(N);
            else                                           ---   M<N
              E3(K) := E3(K)+InvRadPower(2*M)*A1(M)*E2(N); --- norms
            end if;
            K := N+M;                                      --- freq sum
            if K <= EDeg then
              E3(K) := E3(K)+A1(M)*E2(N); --- norms
            else
              E3L := E3L+A1(M)*E2(N); --- norms
            end if;
          end if;
        end loop;
      end if;
    end loop;
  end CEProd2;

  procedure EEProd2A(E1,E2: in Errs1; E3: in out Errs1) is
    --- add twice E1*E2
    --- slower version, more accurate
    --- for efficiency, check first whether E1=0 or E2=0
    E3F: Radius renames E3(0);
    E3L: Radius renames E3(EDeg);
    K: Integer;
  begin
    for M in EFreq loop
      if E1(M) /= Zero then
        for N in EFreq loop
          if E2(N) /= Zero then
            K := N-M;                                      --- freq diff
            if K<0 then                                    ---   M>N
              E3F := E3F+InvRadPower(2*N)*E1(M)*E2(N);
            else                                           ---   M<N
              E3F := E3F+InvRadPower(2*M)*E1(M)*E2(N);
            end if;
            K := N+M;                                      --- freq sum
            if K <= EDeg then
              E3(K) := E3(K)+E1(M)*E2(N);
            else
              E3L := E3L+E1(M)*E2(N);
            end if;
          end if;
        end loop;
      end if;
    end loop;
  end EEProd2A;

  procedure EEProd2B(E1,E2: in Errs1; E3: in out Errs1) is
    --- add twice E1*E2
    --- faster version, less accurate
    --- for efficiency, check first whether E1=0 or E2=0
    ESum1,ESum2: Flt := Zero;
  begin
    for K in reverse EFreq loop
      ESum1 := ESum1+E1(K);
      ESum2 := ESum2+E2(K);
    end loop;
    E3(0) := E3(0)+Two*(ESum1*ESum2);
  end EEProd2B;

  procedure Norms(P: in Parity; C: in FPoly1; A: out CNorms) is
  begin
    if P=0 then A(0) := MaxNorm(C(0)); else A(0) := Zero; end if;
    for I in 1 .. FDeg loop A(I) := RadPower(I)*MaxNorm(C(I)); end loop;
  end Norms;

  procedure Product(P1,P2: in CosSin1; P3: in out CosSin1) is
    --- for FDeg>0
    C1: FPoly1 renames P1.C;
    E1: Errs1 renames P1.E;
    ZC1: constant Boolean := Is_Zero(P1.P,C1);
    ZE1: constant Boolean := Is_Zero(E1);
  begin
    FillZero(P3);
    P3.P := (P1.P+P2.P) mod 2;
    if ZC1 and then ZE1 then return; end if; -- P1=0
    declare
      C2: FPoly1 renames P2.C;
      E2: Errs1 renames P2.E;
      ZC2: constant Boolean := Is_Zero(P2.P,C2);
      ZE2: constant Boolean := Is_Zero(E2);
    begin
      if ZC2 and then ZE2 then return; end if; -- P2=0
      declare
        E3: Errs1 renames P3.E;
        A1,A2: CNorms;
      begin
        if not ZC1 then Norms(P1.P,C1,A1); end if;
        if not ZC2 then Norms(P2.P,C2,A2); end if;
        if not (ZC1 or else ZC2) then
          CCProd2(P1.P,P2.P,C1,C2,A1,A2,P3);
        end if;
        if not (ZC1 or else ZE2) then
          CEProd2(P1.P,A1,E2,E3);
        end if;
        if not (ZC2 or else ZE1) then
          CEProd2(P2.P,A2,E1,E3);
        end if;
        if not (ZE1 or else ZE2) then
          EEProd2(E1,E2,E3);
        end if;
        if Not_Trunc and then P3.P=1 then LiftE0(E3); end if;
      end;
    end;
    Mult(SHalf,P3);
  end Product;

  procedure Prod(P1,P2: in CosSin1; P3: in out CosSin1) is
  begin
    if FDeg=0 then
      P3.P := 0;
      Prod(P1.C(0),P2.C(0),P3.C(0));
    else
      Product(P1,P2,P3);
    end if;
  end Prod;

  function "*"(P1,P2: CosSin1) return CosSin1 is
    P3: CosSin1;
  begin
    Prod(P1,P2,P3);
    return P3;
  end "*";

  procedure AddProd(P1,P2: in CosSin1; P3: in out CosSin1) is
  begin
    if FDeg=0 then
      declare
        Tmp: Scalar;
      begin
        AddProd(P1.C(0),P2.C(0),P3.C(0),Tmp);
      end;
    else
      declare
        Tmp: LP_Pointer;
      begin
        PPool.Allocate(Tmp);
        Prod(P1,P2,Tmp.all.Data);
        Add(Tmp.all.Data,P3);
        PPool.Recycle(Tmp);
      end;
    end if;
  end AddProd;

  procedure AddProd(P1,P2: in CosSin1; P3,Tmp: in out CosSin1) is
  begin
    if FDeg=0 then
      AddProd(P1.C(0),P2.C(0),P3.C(0),Tmp.C(0));
    else
      Prod(P1,P2,Tmp);
      Add(Tmp,P3);
    end if;
  end AddProd;

  procedure SubProd(P1,P2: in CosSin1; P3: in out CosSin1) is
  begin
    if FDeg=0 then
      declare
        Tmp: Scalar;
      begin
        SubProd(P1.C(0),P2.C(0),P3.C(0),Tmp);
      end;
    else
      declare
        Tmp: LP_Pointer;
      begin
        PPool.Allocate(Tmp);
        Prod(P1,P2,Tmp.all.Data);
        Sub(Tmp.all.Data,P3);
        PPool.Recycle(Tmp);
      end;
    end if;
  end SubProd;

  procedure SubProd(P1,P2: in CosSin1; P3,Tmp: in out CosSin1) is
  begin
    if FDeg=0 then
      SubProd(P1.C(0),P2.C(0),P3.C(0),Tmp.C(0));
    else
      Prod(P1,P2,Tmp);
      Sub(Tmp,P3);
    end if;
  end SubProd;

  procedure Div(R: in Flt; P: in out CosSin1) is
    C: FPoly1 renames P.C;
    Tmp: Scalar;
  begin
    if FDeg=0 then
      Div(R,C(0),Tmp);
    else
      for I in P.P .. FDeg loop Div(R,C(I),Tmp); end loop;
      if Not_Trunc then
        declare
          A: constant Flt := abs(R);
          E: Errs1 renames P.E;
        begin
          for I in EFreq loop E(I) := E(I)/A; end loop;
        end;
      end if;
    end if;
  end Div;

  procedure Div(S: in Scalar; P: in out CosSin1) is
  begin
    if FDeg=0 then
      Div(S,P.C(0));
    else
      declare
        InvS: Scalar;
      begin
        Inv(S,InvS);
        Mult(InvS,P);
      end;
    end if;
  end Div;

  procedure Quot(P1: in CosSin1; R: in Flt; P2: in out CosSin1) is
  begin
    if FDeg=0 then
      P2.P := 0;
      Quot(P1.C(0),R,P2.C(0));
    else
      P2.P := P1.P;
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
      begin
        for I in P1.P .. FDeg loop Quot(C1(I),R,C2(I)); end loop;
      end;
      if Not_Trunc then
        declare
          A: constant Flt := abs(R);
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
        begin
          for I in EFreq loop E2(I) := E1(I)/A; end loop;
        end;
      end if;
    end if;
  end Quot;

  function "/"(P: CosSin1; R: Flt) return CosSin1 is
    Q: CosSin1;
  begin
    Quot(P,R,Q);
    return Q;
  end "/";

  ------------------------------------------------

  function Get_Precision(P: CosSin1) return Positive is
  begin
    return Get_Precision(P.C(P.P));
  end Get_Precision;

  procedure Assign(I: in Integer; P: in out CosSin1) is
  begin
    P.P := 0;
    if FDeg>0 then FillZero(P); end if;
    Assign(I,P.C(0));
  end Assign;

  procedure Assign(Q: in Rational; P: in out CosSin1) is
  begin
    P.P := 0;
    if FDeg>0 then FillZero(P); end if;
    Assign(Q,P.C(0));
  end Assign;

  procedure Assign(R: in Flt; P: in out CosSin1) is
  begin
    P.P := 0;
    if FDeg>0 then FillZero(P); end if;
    Assign(R,P.C(0));
  end Assign;

  procedure Assign(S: in Scalar; P: in out CosSin1) is
  begin
    P.P := 0;
    if FDeg>0 then FillZero(P); end if;
    Copy(S,P.C(0));
  end Assign;

  procedure Show(N: in String; P: in CosSin1; Cut: in Radius; Hide0: in Boolean := True) is
    NotHide0: constant Boolean := not Hide0;
    CosSinPar: constant String := CosSin(P.P);
    C: FPoly1 renames P.C;
    E: Errs1 renames P.E;
  begin
    if MaxNorm(P)<=Cut then
      null;
    elsif FDeg=0 then
      if NotHide0 or else MaxNorm(C(0))>Cut then
        Show1(N,C(0));
      end if;
    elsif not IsZero(P) then
      for I in P.P .. FDeg loop
        if NotHide0 or else MaxNorm(C(I))>Cut then
          Show1(N & " " & CosSinPar & Strng(I,4),C(I));
        end if;
      end loop;
      if not IsPoly(P) then
        for I in EFreq loop
          if NotHide0 or else E(I)>Cut  then
            Show1(N & " " & "err" & Strng(I,4),E(I));
          end if;
        end loop;
      end if;
    elsif NotHide0 then
      Show0(N & " 0");
    end if;
  end Show;

  procedure Show1(N: in String; P: in CosSin1; Hide0: in Boolean := True) is
  begin
    Show(N,P,Zero,Hide0);
  end Show1;

  procedure Show2(N: in String; P1,P2: in CosSin1; Hide0: in Boolean := True) is
    CosSinPar: constant String := CosSin(P1.P);
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    E1: Errs1 renames P1.E;
    E2: Errs1 renames P2.E;
  begin
    if FDeg=0 then
      Show2(N,C1(0),C2(0),Hide0);
    elsif not (IsZero(P1) and then IsZero(P2)) then
      if P1.P/=P2.P then raise Not_Implemented; end if;
      for I in P1.P .. FDeg loop
        if not (IsZero(C1(I)) and then IsZero(C2(I)) and then Hide0) then
          Show2(N & " " & CosSinPar & Strng(I,4),C1(I),C2(I));
        end if;
      end loop;
      if not (IsPoly(P1) and then IsPoly(P2)) then
        for I in EFreq loop
          if not (E1(I)=Zero and then E2(I)=Zero and then Hide0) then
            Show2(N & " " & "err" & Strng(I,4),E1(I),E2(I));
          end if;
        end loop;
      end if;
    elsif not Hide0 then
      Show0(N & " 0 0");
    end if;
  end Show2;

  function QuasiDeg(C: FPoly1; P: Parity := 0) return Natural is
  begin
    for I in reverse P .. FDeg loop
      if not IsZero(C(I)) then return I; end if;
    end loop;
    return 0;
  end QuasiDeg;

  function QuasiDeg(E: Errs1) return Natural is
  begin
    if Not_Trunc then
      for I in reverse 1 .. EDeg loop
        if E(I) /= Zero then return I; end if;
      end loop;
    end if;
    return 0;
  end QuasiDeg;

  procedure Put(F: in File_Type; P: in CosSin1; Simple: in Boolean := False) is
    --- if Simple, then FDeg=0 CosSin1 is written like a Scalar
    Decimal: constant Boolean := False;
    C: FPoly1 renames P.C;
    E: Errs1 renames P.E;
    DC,DE: Integer;
  begin
    if FDeg=0 then
      if Simple then
        Put(F,C(0),Decimal);
      else
        Ints.Put(F,0);         --- Parity
        Ints.Put(F,0);         --- degree
        Put(F,C(0),Decimal);   --- coeff
        Ints.Put(F,0);         --- error deg
        Put(F,Zero,Decimal);   --- zero error
      end if;
    else
      DC := QuasiDeg(C,P.P);
      Put(F,P.P);              --- Parity
      Put(F,DC);               --- degree
      for I in P.P .. DC loop
        Put(F,C(I),Decimal);   --- coeffs
      end loop;
      if Trunc then
        Ints.Put(F,0);         --- error deg
        Put(F,Zero,Decimal);   --- zero error
      else
        DE := QuasiDeg(E);
        Put(F,DE);             --- error deg
        for I in 0 .. DE loop
          Put(F,E(I),Decimal); --- errros
        end loop;
      end if;
    end if;
  end Put;

  procedure Get(F: in File_Type; P: in out CosSin1; Simple: in Boolean := False) is
    Decimal: constant Boolean := False;
    DC,DE: Integer := 0;
    R: Flt := Zero;
    S: Scalar;
  begin
    if Simple then
      P.P := 0;
      if FDeg>0 then FillZero(P); end if;
      Get(F,P.C(0),Decimal);
    else
      FillZero(P);
      Get(F,P.P);
      Get(F,DC);
      for I in P.P .. DC loop
        Get(F,S,Decimal);
        AddCoeff(I,S,P);
      end loop;
      Get(F,DE);
      for I in 0 .. DE loop
        Get(F,R,Decimal);
        AddErr(I,R,P);
      end loop;
      if Not_Trunc and then P.P=1 then LiftE0(P.E); end if;
    end if;
  end Get;

  procedure Write(FileName: in String; P: in CosSin1; Simple: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,P,Simple);
    Close(F);
  end Write;

  procedure Read(FileName: in String; P: in out CosSin1; Simple: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,P,Simple);
    Close(F);
  end Read;

  -----------------------------------------

  procedure Simple_Random(P: in out CosSin1) is
    --- assumes P.P has been set
    Par: constant Parity := P.P;
  begin
    Simple_Random(Rad,Par,P);
  end Simple_Random;

  procedure Simple_Random(R: in Radius; Par: in Parity; P: in out CosSin1) is
    C: FPoly1 renames P.C;
    Fac: Flt := One;
  begin
    P.P := Par;
    Simple_Random(C(0));
    if FDeg>0 then
      for I in 1 .. FDeg loop
        Fac := Fac/R;
        Assign(Fac*Simple_Random,C(I));
      end loop;
      if Not_Trunc then P.E := (others => Zero); end if;
    end if;
  end Simple_Random;

  function MaxNorm(P: CosSin1) return Radius is
  begin
    if FDeg=0 then return MaxNorm(P.C(0)); end if;
    declare
      C: FPoly1 renames P.C;
      E: Errs1 renames P.E;
      S: Flt := Zero;
    begin
      if Not_Trunc then
        for I in reverse EFreq loop S := S+E(I); end loop;
      end if;
      for I in reverse P.P .. FDeg loop
        S := S+RadPower(I)*MaxNorm(C(I));
      end loop;
      return S;
    end;
  end MaxNorm;

  procedure Norm(P: in CosSin1; S: in out Scalar) is
  begin
    if FDeg=0 then
      Norm(P.C(0),S);
    else
      declare
        C: FPoly1 renames P.C;
        Tmp: Scalar;
      begin
        SetZero(S);
        for I in reverse 1 .. FDeg loop
          Norm(C(I),Tmp);
          Add(Tmp,S);
          Mult(Rad,S,Tmp);
        end loop;
        if P.P=0 then
          Norm(C(0),Tmp);
          Add(Tmp,S);
        end if;
        if Not_Trunc then
          declare
            E: Errs1 renames P.E;
            R: Flt := Zero;
          begin
            for I in reverse EFreq loop R := R+E(I); end loop;
            BallAt0(R,Tmp);
          end;
          Add(Tmp,S);
        end if;
      end;
    end if;
  end Norm;

  function Norm(P: CosSin1) return Scalar is
    S: Scalar;
  begin
    Norm(P,S);
    return S;
  end Norm;

  function Inf(P: CosSin1) return Flt is
  begin
    if FDeg>0 then raise Not_Implemented; end if;
    return Inf(P.C(0));
  end Inf;

  function Sup(P: CosSin1) return Flt is
  begin
    if FDeg>0 then raise Not_Implemented; end if;
    return Sup(P.C(0));
  end Sup;

  procedure Average(P: in CosSin1; S: in out Scalar) is
  begin
    if P.P=1 then
      SetZero(S);
    elsif NoErrs then
      Copy(P.C(0),S);
    else
      BallAt0(P.E(0),S);
      Add(P.C(0),S);
    end if;
  end Average;

  function Average(P: CosSin1) return Scalar is
    S: Scalar;
  begin
    Average(P,S);
    return S;
  end Average;

  procedure SetAverage0(P: in out CosSin1) is
  begin
    SetZero(P.C(0));
    if Errs then LiftE0(P.E); end if;
  end SetAverage0;

  procedure Coeff(I: in Natural; P: in CosSin1; S: in out Scalar) is
  begin
    if FDeg=0 then
      if I=0 then Copy(P.C(0),S); else SetZero(S); end if;
    elsif I=0 then
      Average(P,S);
    elsif Trunc then
      if I>FDeg then SetZero(S); else Copy(P.C(I),S); end if;
    else      --- Errs
      declare
        E: Errs1 renames P.E;
        R: Flt := Zero;
      begin
        for K in reverse 0 .. Imin(I,EDeg) loop R := R+E(K); end loop;
        BallAt0(InvRadPower(I)*R,S);
      end;
      if I<=FDeg then Add(P.C(I),S); end if;
    end if;
  end Coeff;

  function Coeff(I: Natural; P: CosSin1) return Scalar is
    S: Scalar;
  begin
    Coeff(I,P,S);
    return S;
  end Coeff;

  procedure AddCoeff(I: in Natural; S: in Scalar; P: in out CosSin1) is
  begin
    if FDeg=0 then
      if I/=0 then raise Undefined; end if;
      Add(S,P.C(0));
    elsif not IsZero(S) then
      if I <= FDeg then
        if I<P.P then raise Undefined; end if;
        Add(S,P.C(I));
      elsif Not_Trunc then
        AddErr(I,MaxNorm(S),P);
      end if;
    end if;
  end AddCoeff;

  procedure AddErr(I: in Natural; R: in Radius; P: in out CosSin1) is
    --- higher order erors, R is a norm
  begin
    if Not_Trunc then
      if FDeg=0 then raise Not_Implemented; end if;
      declare
        E: Errs1 renames P.E;
        EI: Radius renames E(IMin(I,EDeg));
      begin
        EI := EI+R;
        if P.P=1 then LiftE0(E); end if;
      end;
    end if;
  end AddErr;

  procedure AddBall(R: in Radius; P: in out CosSin1) is
    --- R is a norm
  begin
    if Trunc or else R=Zero then
      null;
    elsif FDeg=0 then
      if P.P=1 then raise Inconsistent_Data; end if;
      declare
        S: Scalar;
      begin
        BallAt0(R,S);
        Add(S,P.C(0));
      end;
    else
      declare
        E: Errs1 renames P.E;
      begin
        E(0) := E(0)+R;
        if P.P=1 then LiftE0(E); end if;
      end;
    end if;
  end AddBall;

  function MaxVal(E: Errs1) return Radius is
    V: Flt := Zero;
  begin
    if Errs then
      for I in reverse EFreq loop
        V := InvRad*V+E(I);
      end loop;
    end if;
    return V;
  end MaxVal;

  procedure ValAt0(P: in CosSin1; S: in out Scalar; D: in Natural; Is_Poly: in Boolean) is
    --- assuming FDeg>0 and D=QuasiDeg(P.C,P.P);
  begin
    if P.P=0 then            -- even
      if Is_Poly then
        SetZero(S);
      else
        BallAt0(MaxVal(P.E),S);
      end if;
      declare
        C: FPoly1 renames P.C;
      begin
        for I in reverse 0 .. D loop Add(C(I),S); end loop;
      end;
    else                     -- odd
      SetZero(S);
    end if;
  end ValAt0;

  procedure ValAt0(P: in CosSin1; S: in out Scalar) is
  begin
    ValAt0(P,S,QuasiDeg(P.C,P.P),IsPoly(P));
  end ValAt0;

  procedure Val(P: in CosSin1; X: in Scalar; Y: in out Scalar; Tmp1,Tmp2: in out FPoly1) is
  begin
    if FDeg=0 then
      Copy(P.C(0),Y);
    else                        -- FDeg>0
      declare
        C: FPoly1 renames P.C;
        D: constant Natural := QuasiDeg(C,P.P);
        Is_Poly: constant Boolean := IsPoly(P);
      begin
        if IsZero(X) then        -- zero X
          ValAt0(P,Y,D,Is_Poly);
        else                     -- nonzero X
          if Is_Poly then        --   no errors
            SetZero(Y);
          else                   --   evaluate errorrs
            BallAt0(MaxVal(P.E),Y);
          end if;
          if D=0 then
            Add(C(0),Y);
          else                   -- D>0
            declare
              CP: FPoly1 renames Tmp1;
              SP: FPoly1 renames Tmp2;
              S: Scalar renames SP(0); -- useless anyhow
            begin
              CSPowers(X,CP,SP);
              if P.P=0 then
                for I in reverse 1 .. D loop
                  Prod(C(I),CP(I),S);
                  Add(S,Y);
                end loop;
                Add(C(0),Y);
              else
                for I in reverse 1 .. D loop
                  Prod(C(I),SP(I),S);
                  Add(S,Y);
                end loop;
              end if;
            end;
          end if; -- D
        end if; -- X
      end;
    end if; -- FDeg
  end Val;

  procedure Val(P: in CosSin1; X: in Scalar; Y: in out Scalar) is
  begin
    if FDeg=0 then
      Copy(P.C(0),Y);
    elsif IsZero(X) then
      ValAt0(P,Y);
    else
      declare
        Tmp1,Tmp2: LP_Pointer;
      begin
        PPool.Allocate(Tmp1);
        PPool.Allocate(Tmp2);
        Val(P,X,Y,Tmp1.all.Data.C,Tmp2.all.Data.C);
        PPool.Recycle(Tmp1);
        PPool.Recycle(Tmp2);
      end;
    end if;
  end Val;

  function Val(P: CosSin1; S: Scalar) return Scalar is
    V: Scalar;
  begin
    Val(P,S,V);
    return V;
  end Val;

  procedure TranslatePi(P: in out CosSin1) is
    --- C(odd) := -C(odd)
  begin
    if FDeg>0 then
      declare
        C: FPoly1 renames P.C;
        I: Integer := 1;
      begin
        while I <= FDeg loop Neg(C(I)); I := I+2; end loop;
      end;
    end if;
  end TranslatePi;

  procedure TranslatePi(P1: in CosSin1; P2: in out CosSin1) is
  begin
    P2.P := P1.P;
    if FDeg=0 then
      Copy(P1.C(0),P2.C(0));
    else
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
        Odd: Boolean := (P1.P=1);
      begin
        for I in P1.P .. FDeg loop
          if Odd then Neg(C1(I),C2(I)); else Copy(C1(I),C2(I)); end if;
          Odd := not Odd;
        end loop;
      end;
      if Not_Trunc then P2.E := P1.E; end if;
    end if;
  end TranslatePi;

  procedure NegTranslatePi(P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    Odd: Boolean := (P1.P=1);
  begin
    P2.P := P1.P;
    if FDeg>0 then
      for I in P1.P .. FDeg loop
        if Odd then Copy(C1(I),C2(I)); else Neg(C1(I),C2(I)); end if;
        Odd := not Odd;
      end loop;
      if Not_Trunc then P2.E := P1.E; end if;
    elsif IsZero(C1(0)) then
      SetZero(C2(0));
    else
      raise Undefined;
    end if;
  end NegTranslatePi;

  procedure SymmetrizePi(SP: in Parity; P: in out CosSin1; Check: in Boolean := False) is
    --- SP=0: symmetric, C(odd) := 0
    --- SP=1: antisymmetric, C(even) := 0
    I: Integer;
  begin
    if FDeg=0 then
      if SP=1 then SetZero(P.C(0)); end if;
    else
      declare
        C: FPoly1 renames P.C;
      begin
        I := 1-SP;
        while I <= FDeg loop
          if not IsZero(C(I)) then
            if Check then raise Sorry; end if;
            SetZero(C(I));
          end if;
          I := I+2;
        end loop;
      end;
      if Not_Trunc then
        declare
          E: Errs1 renames P.E;
        begin
          if P.P=1 then LiftE0(E); end if;
          I := 1-SP;
          while I<EDeg loop
            E(I+1) := E(I+1)+E(I);
            E(I) := Zero;
            I := I+2;
          end loop;
        end;
      end if;
    end if;
  end SymmetrizePi;

  procedure ProjEvenFreq(P: in out CosSin1) is
    --- SymmetrizePi(0,P,False)
    --- C(odd)=0
  begin
    if FDeg>0 then
      declare
        C: FPoly1 renames P.C;
        I: Integer := 1;
      begin
        while I <= FDeg loop SetZero(C(I)); I := I+2; end loop;
      end;
      if Not_Trunc then
        declare
          E: Errs1 renames P.E;
          I: Integer := 1;
        begin
          if P.P=1 then LiftE0(E); end if;
          while I<EDeg loop
            E(I+1) := E(I+1)+E(I);
            E(I) := Zero;
            I := I+2;
          end loop;
        end;
      end if;
    end if;
  end ProjEvenFreq;

  procedure ProjOddFreq(P: in out CosSin1) is
    --- SymmetrizePi(1,P,False)
    --- C(even)=0
  begin
    if FDeg=0 then
      SetZero(P.C(0));
    else
      declare
        C: FPoly1 renames P.C;
        I: Integer := 0;
      begin
        while I <= FDeg loop SetZero(C(I)); I := I+2; end loop;
      end;
      if Not_Trunc then
        declare
          E: Errs1 renames P.E;
          I: Integer := 0;
        begin
          while I<EDeg loop
            E(I+1) := E(I+1)+E(I);
            E(I) := Zero;
            I := I+2;
          end loop;
        end;
      end if;
    end if;
  end ProjOddFreq;

  procedure ProjPar(Par: in Parity; P: in out CosSin1; Check: in Boolean := False) is
  begin
    if P.P/=Par then
      if IsZero(P) then
        P.P := Par;
        if Par=0 then SetZero(P.C(0)); end if;
      elsif Check then
        raise Parity_Error;
      end if;
    elsif Par=1 and then Errs then
      declare
        E: Errs1 renames P.E;
      begin
        E(1) := E(1)+E(0);
        E(0) := Zero;
      end;
    end if;
  end ProjPar;

  procedure ProjConst(P: in out CosSin1) is
  begin
    if FDeg>0 then
      declare
        C: FPoly1 renames P.C;
      begin
        for I in 1-P.P .. FDeg loop SetZero(C(I)); end loop;
      end;
      if Not_Trunc then
        if P.P=0 then
          declare
            E0: constant Radius := P.E(0);
          begin
            P.E := (others => Zero);
            P.E(0) := E0;
          end;
        else
          P.E := (others => Zero);
        end if;
      end if;
    end if;
  end ProjConst;

  procedure Der(P: in out CosSin1) is
    C: FPoly1 renames P.C;
    Tmp: Scalar renames C(0);
  begin
    if FDeg=0 then
      SetZero(C(0));
    else
      if not IsPoly(P) then raise Not_Implemented; end if; -- cannot decrease Rad
      if P.P=0 then
        P.P := 1;
        Neg(C(1));
        for I in 2 .. QuasiDeg(C,1) loop
          Mult(Flt(-I),C(I),Tmp);
        end loop;
      else
        P.P := 0;
        SetZero(C(0));
        for I in 2 .. QuasiDeg(C,1) loop
          Mult(Flt(I),C(I),Tmp);
        end loop;
      end if;
    end if;
  end Der;

  procedure Der(P1: in CosSin1; P2: in out CosSin1) is
  begin
    Copy(P1,P2);
    Der(P2);
  end Der;

  procedure AntiDer(P: in out CosSin1) is
    C: FPoly1 renames P.C;
    Tmp: Scalar renames C(0);
  begin
    if not IsZero(Tmp) then raise Undefined; end if;
    if FDeg>0 then
      if P.P=0 then
        P.P := 1;
        for I in 2 .. QuasiDeg(C,1) loop Div(Flt(I),C(I),Tmp); end loop;
      else
        P.P := 0;
        Neg(C(1));
        for I in 2 .. QuasiDeg(C,1) loop Div(Flt(-I),C(I),Tmp); end loop;
      end if;
      SetZero(Tmp);
      if not IsPoly(P) then
        declare
          E: Errs1 renames P.E;
        begin
          LiftE0(E);
          for I in 2 .. EDeg loop
            E(I) := E(I)/Flt(I);
          end loop;
        end;
      end if;
    end if;
  end AntiDer;

  procedure AntiDer(P1: in CosSin1; P2: in out CosSin1) is
  begin
    Copy(P1,P2);
    AntiDer(P2);
  end AntiDer;

  procedure Integral(P: in CosSin1; X0,X: in Scalar; Y: in out Scalar) is
    Tmp: LP_Pointer;
    Y0,A: Scalar;
  begin
    if FDeg=0 then
      SetZero(Y);
    else
      PPool.Allocate(Tmp);
      declare
        Q: CosSin1 renames Tmp.all.Data;
      begin
        Copy(P,Q);
        SetAverage0(Q);
        AntiDer(Q);
        Val(Q,X0,Y0);
        Val(Q,X,Y);
      end;
      PPool.Recycle(Tmp);
      Sub(Y0,Y);
    end if;
    Average(P,A);
    if not IsZero(A) then
      Diff(X,X0,Y0);
      Mult(A,Y0);
      Add(Y0,Y);
    end if;
  end Integral;

  procedure InvHelm(A,B: in Scalar; Del: in Radius; P: in out CosSin1; OddFreq: in Boolean := True) is
    --- inverse of Helmholtz operator B^2+(A*D)^2 with B,A positive
    --- assuming |B-A*N|>=Del for when B>A*EDeg , then (B-A*N)(B+A*N) >= Del*(B+A*N)
    Par: constant Parity := P.P; --- not need to check (in loops) whether P.P changed
    C: FPoly1 renames P.C;
    B2: Scalar;
  begin
    if Inf(B)<Zero then raise Not_Implemented; end if;
    Prod(B,B,B2);
    if Par=0 and then (not IsZero(C(0))) then
      Div(B2,C(0));
    end if;
    if FDeg>0 then
      if Inf(A)<Zero then raise Not_Implemented; end if;
      declare
        A2,S: Scalar;
        DN: constant Positive := Choose(OddFreq,2,1);
        N: Integer := 1;
      begin
        Prod(A,A,A2);
        while N<=FDeg loop
          if not IsZero(C(N)) then
            Prod(Flt(-(N*N)),A2,S);
            Add(B2,S);                 --- B^2-(A*N)^2
            Div(S,C(N));
          end if;
          N := N+DN;
        end loop;
        if not IsPoly(P) then
          declare
            E: Errs1 renames P.E;
            NA: Scalar renames A2;
            Dist,MinDist: Radius;
            LastN: constant Positive := Choose(OddFreq,EDegOdd,EDeg);
          begin
            if Par=1 then E(1) := E(1)+E(0); E(0) := Zero; end if;
            N := LastN;
            if N<EDeg then E(N) := E(N)+E(EDeg); E(EDeg) := Zero; end if; -- not optimal
            Prod(Flt(N),A,NA);
            Diff(NA,B,S);                            --- S=EDeg*A-B
            if Inf(S)<Zero then                      --- use Del if S<0
              while N>=Par loop                      --- now N<EDeg
                if E(N)/=Zero then
                  Sum(NA,B,S);
                  E(N) := (E(N)/Del)/Inf(S);
                end if;
                N := N-DN;
              end loop;
            else                                     --- Del not needed
              N := LastN;
              MinDist := Radius'Last;
              while N>=Par loop
                Prod(Flt(N),A,NA);
                Diff(NA,B,S);                                 --- N*A-B
                if Approx(S)<Zero then Neg(S); end if;
                Dist := Inf(S);                               --- fails if Inf(S)<0
                if Dist<MinDist then MinDist := Dist; end if;
                if E(N)/=Zero then
                  Sum(NA,B,S);                                --- N*A+B
                  E(N) := (E(N)/MinDist)/Inf(S);
                end if;
                N := N-DN;
              end loop;
            end if;
          end;
        end if;
      end;
    end if;
  end InvHelm;

  ----------------------------- Fourier1

  procedure AddProd(S: in Scalar; P1: in CosSin1; P2: in out Fourier1) is
  begin
    if P1.P=0 then
      AddProd(S,P1,P2.E);
    else
      AddProd(S,P1,P2.O);
    end if;
  end AddProd;

  procedure Val(P: in Fourier1; X: in Scalar; Y: in out Scalar) is
  begin
    if FDeg=0 then
      Copy(P.E.C(0),Y);
    elsif IsZero(P.O) then
      Val(P.E,X,Y);
    elsif IsZero(P.E) then
      Val(P.O,X,Y);
    else
      declare
        Tmp1,Tmp2: LP_Pointer;
        S: Scalar;
      begin
        PPool.Allocate(Tmp1);
        PPool.Allocate(Tmp2);
        Val(P.E,X,Y,Tmp1.all.Data.C,Tmp2.all.Data.C);
        Val(P.O,X,S,Tmp1.all.Data.C,Tmp2.all.Data.C);
        PPool.Recycle(Tmp1);
        PPool.Recycle(Tmp2);
        Add(S,Y);
      end;
    end if;
  end Val;

  ----------------------------- numeric

  procedure NumTrunc(Siz: in Natural; P: in out CosSin1) is
    C: FPoly1 renames P.C;
  begin
    if FDeg=0 then
      Center(C(0));
    elsif Siz>=FDeg then
      for I in P.P .. FDeg loop Center(C(I)); end loop;
      if Not_Trunc then P.E := (others => Zero); end if;
    else
      for I in P.P .. Siz loop Center(C(I)); end loop;
      for I in Siz+1 .. FDeg loop SetZero(C(I)); end loop;
      if Not_Trunc then P.E := (others => Zero); end if;
    end if;
  end NumTrunc;

  procedure NumTrunc(Siz: in Natural; P1: in CosSin1; P2: in out CosSin1) is
  begin
    Copy(P1,P2);
    NumTrunc(Siz,P2);
  end NumTrunc;

  function NumVal(P: CosSin1; T: Flt) return Flt is
  begin
    if FDeg=0 then return Approx(P.C(0)); end if;
    declare
      C: FPoly1 renames P.C;
      D: constant Integer := QuasiDeg(C,P.P);
      S: Flt := Zero;
      use Flt_EF;
    begin
      if P.P=1 then
        if T/=Zero then
          for I in reverse 1 .. D loop S := S+Approx(C(I))*Sin(Flt(I)*T); end loop;
        end if;
      elsif T=Zero then
        for I in reverse 0 .. D loop S := S+Approx(C(I)); end loop;
      else
        for I in reverse 0 .. D loop S := S+Approx(C(I))*Cos(Flt(I)*T); end loop;
      end if;
      return S;
    end;
  end NumVal;

  function NumVal(P: Fourier1; T: Flt) return Flt is
  begin
    if FDeg=0 then return NumVal(P.E,T); end if;
    return NumVal(P.E,T)+NumVal(P.O,T);
  end NumVal;

  procedure PutPlot(F: in File_Type; P: in CosSin1; Steps: in Positive := 64; PutX: in Boolean := False) is
    Digs: constant Positive := 4;
    Pi2: constant := 6.28318_53071_79586_47692_52867_66559;
    X,Y: Flt;
    use FLt_IO;
  begin
    X := Zero;
    Y := NumVal(P,X);
    if PutX then Put(F,X,3,Digs,4); end if;
    Put(F,Y,3,Digs,4);
    New_Line(F);
    if FDeg>0 then
      for N in 1 .. Steps loop
        X := Pi2*Flt(N)/Flt(Steps);
        Y := NumVal(P,X);
        if PutX then Put(F,X,3,Digs,4); end if;
        Put(F,Y,3,Digs,4);
        New_Line(F);
      end loop;
    end if;
  end PutPlot;

  function GuessRho(P: CosSin1) return Radius is
  begin
    if FDeg<5 then return Radius'Last; end if;
    declare
      DMin: constant Integer := FDeg/5;
      DMax: constant Integer := (4*FDeg)/5;
      C: FPoly1 renames P.C;
      I1: Integer := DMin;
      I2: Integer := DMax;
      R1,R2: Flt;
      use Flt_EF;
    begin
      loop
        R1 := MaxNorm(C(I1));
        exit when R1/=Zero;
        I1 := I1+1;
      end loop;
      loop
        R2 := MaxNorm(C(I2));
        exit when R2/=Zero;
        I2 := I2-1;
      end loop;
      return Exp(Log(R1/R2)/Flt(I2-I1));
    end;
  end GuessRho;

end TCosSins1;
