with Globals, Strings, Protected_Counters;
use Globals, Strings, Protected_Counters;

pragma Elaborate_All (Globals,Strings,Protected_Counters);

package body TSpherics is

  procedure Show_Param(All_Param: in Boolean := True) is
  begin
    Show_C1_Param(All_Param);
    Show1("TS:           Size = ",Size);
    Show1("TS:            Rad =",Rho1);
  end Show_Param;

  --------------------------------------
  --- associated Legendre functions ----
  --------------------------------------

  procedure ALegend_Vals(C,S: in Scalar; V: in out SArray) is
    --- values V(L,M)=P(c) if c in [-1,1] and s=Sqrt(1-c*c)
    --- L<M is not used - scratch space:
    Num: Scalar renames V(0,1);
    Den: Scalar renames V(0,2);
    Fac: Scalar renames V(0,3);
    T:   Scalar renames V(0,4);
    Tmp: Scalar renames V(0,5);    --- assuming Size >= 5
    L,M1: Integer;
  begin
    Copy(SOne,V(0,0));             --- have V(0,0),
                                   --- need V(1,1),V(2,2),...,V(Size,Size)
    for M in 0 .. Size-1 loop      --- need V(1,0),V(2,1),...,V(Size,Size-1)
      M1 := M+1;
      Assign(Flt(M+M1),Num);       --- 2M+1
      Assign(Flt(2*M1),Den);       --- 2M+2
      Quot(Num,Den,Tmp);
      Sqrt(Tmp,Fac);
      Neg(Fac);
      Mult(S,Fac);                 --- Fac=-Sqrt{(2M+1)/(2M+2)}*S
      Prod(Fac,V(M,M),V(M1,M1));   --- V(M+1,M+1)=Fac*V(M,M)
      Sqrt(Num,Fac);
      Mult(C,Fac);                 --- Fac=Sqrt{2M+1}*C
      Prod(Fac,V(M,M),V(M1,M));    --- V(M+1,M)=Fac*V(M,M)
    end loop;
    --- have V(M,M),V(M+1,M)
    for M in 0 .. Size-2 loop      --- need V(M+2,M),V(M+3,M),...,V(Size,M)
      for L1 in M+2 .. Size loop   --- compute V(L1,M)
        L := L1-1;
        Assign(Flt((L+M)*(L-M)),Tmp);
        Sqrt(Tmp,T);
        Prod(T,V(L-1,M),Fac);      --- Fac = Sqrt((L+M)(L-M)}*V(L-1,M)
        Prod(Flt(L+L1),C,Tmp);     ---       (2L+1)*C
        Prod(Tmp,V(L,M),Num);
        Sub(Fac,Num);              --- Num = (2L+1)*C*V(L,M)-Fac
        Assign(Flt((L1+M)*(L1-M)),Tmp);
        Sqrt(Tmp,Den);             --- Den = Sqrt{(L+1+M)(L+1-M)}
        Quot(Num,Den,V(L1,M));     --- V(L+1,M) = Num/Den
      end loop;
    end loop;
  end ALegend_Vals;

  procedure ALegend_Vals(C: in Scalar; V: in out SArray) is
    --- values V(L,M)=P(c) if c in [-1,1]
    --- L<M is not used - scratch space:
    CC: Scalar renames V(0,1);
    SS: Scalar renames V(0,2);
    S: Scalar;
  begin
    Prod(C,C,CC);
    Diff(SOne,CC,SS);
    Sqrt(SS,S);          -- S := Sqrt{1-C*C}
    ALegend_Vals(C,S,V);
  end ALegend_Vals;

  procedure ALegendVals(Theta: in Scalar; V: in out SArray) is
    --- values V(L,M)=P(cos(theta)) with theta real
    C,S: Scalar;
  begin
    Cos(Theta,C);
    Sin(Theta,S);
    ALegend_Vals(C,S,V);
  end ALegendVals;

  ------------------------------------
  --- CArray and EArray procedures ---
  ------------------------------------

  function IsZero(C: CArray) return Boolean is
  begin
    for I in CRange loop
      for J in CRange loop
        if not IsZero(C(I,J)) then return False; end if;
      end loop;
    end loop;
    return True;
  end IsZero;

  function IsZero(E: EArray) return Boolean is
  begin
    if Trunc then return True; end if;
    for L in ERange loop
      if E(L)/=Zero then return False; end if;
    end loop;
    return True;
  end IsZero;

  function EffLast(C: CArray) return Integer is
  begin
    for L in reverse CRange loop
      if not IsZero(C(L,L)) then return L; end if;
      for K in 0 .. L-1 loop
        if not IsZero(C(L,K)) then return L; end if;
        if not IsZero(C(K,L)) then return L; end if;
      end loop;
    end loop;
    return -1;
  end EffLast;

  function EffLast(E: EArray) return Integer is
  begin
    if Trunc then return -1; end if;
    for L in reverse ERange loop
      if E(L)/=Zero then return L; end if;
    end loop;
    return -1;
  end EffLast;

  function MaxNorm(C: CArray) return Radius is
    R: Radius;
    S: Radius := Zero;
  begin
    for L in reverse 1 .. Size loop
      R := MaxNorm(C(L,L));        -- M=0
      for K in 0 .. L-1 loop       -- K=L-M
        R := R+MaxNorm(C(L,K))+MaxNorm(C(K,L));
      end loop;
      S := Rho2*S+R;
    end loop;
    S := Rho2*S+MaxNorm(C(0,0));
    return S;
  end MaxNorm;

  function Norm(E: EArray) return Radius is
    S: Radius := Zero;
  begin
    for L in reverse ERange loop S := S+E(L); end loop;
    return S;
  end Norm;

  --------------------------
  --- Spheric procedures ---
  --------------------------

  function "="(F1,F2: Spheric) return Boolean is
    --- strong equality
  begin
    if Not_Trunc then
      declare
        E1: EArray renames F1.E;
        E2: EArray renames F2.E;
      begin
        for L in ERange loop
          if E1(L) /= E2(L) then return False; end if;
        end loop;
      end;
    end if;
    declare
      C1: CArray renames F1.C;
      C2: CArray renames F2.C;
    begin
      for I in CRange loop
        for J in CRange loop
          if C1(I,J) /= C2(I,J) then return False; end if;
        end loop;
      end loop;
    end;
    return True;
  end "=";

  procedure FillZero(F: in out Spheric) is
    --- keep parities
    C: CArray renames F.C;
  begin
    for I in CRange loop
      for J in CRange loop FillZero(C(I,J)); end loop;
    end loop;
    if Not_Trunc then F.E := (others => Zero); end if;
  end FillZero;

  procedure SetZero(F: in out Spheric) is
    --- use default Par and Mixed
    PC: Parity renames Par;                         --- time-parity of cos(m*phi) term
    PS: constant Parity := Choose(Mixed,1-Par,Par); --- time-parity of sin(m*phi) term
    C: CArray renames F.C;
  begin
    for L in CRange loop
      for K in 0 .. L-1 loop
        SetZero(PC,C(L,K));  --- cos(k*phi)
        SetZero(PS,C(K,L));  --- sin(k*phi)
      end loop;
      SetZero(PC,C(L,L));    --- cos(0*phi)
    end loop;
    if Not_Trunc then F.E := (others => Zero); end if;
  end SetZero;

  procedure SetZero(P: in Parity; Mix: in Boolean; F: in out Spheric) is
    PC: Parity renames P;                     --- time-parity of cos(m*phi) term
    PS: constant Parity := Choose(Mix,1-P,P); --- time-parity of sin(m*phi) term
    C: CArray renames F.C;
  begin
    for L in CRange loop
      for K in 0 .. L-1 loop
        SetZero(PC,C(L,K));  --- cos(k*phi)
        SetZero(PS,C(K,L));  --- sin(k*phi)
      end loop;
      SetZero(PC,C(L,L));    --- cos(0*phi)
    end loop;
    if Not_Trunc then F.E := (others => Zero); end if;
  end SetZero;

  procedure SetParity(P: in Parity; Mix: in Boolean; F: in out Spheric) is
    --- Mix=False: parity P under  t     --> -t
    --- Mix=True:  parity P under (t,phi)-->(-t,-phi)
    PC: Parity renames P;                     --- desired time-parity of cos(m*phi) term
    PS: constant Parity := Choose(Mix,1-P,P); --- desired time-parity of sin(m*phi) term
    C: CArray renames F.C;
  begin
    for L in CRange loop
      C(L,L).P := PC;
      for K in 0 .. L-1 loop      --- M=L-K
        C(L,K).P := PC;
        C(K,L).P := PS;
      end loop;
    end loop;
  end SetParity;

  function IsZero(F: Spheric) return Boolean is
  begin
    if Trunc then return IsZero(F.C); end if;
    return IsZero(F.E) and then IsZero(F.C);
  end IsZero;

  function IsSharp(F: Spheric) return Boolean is
  begin
    if Trunc then return True; end if;
    declare
      C: CArray renames F.C;
      E: EArray renames F.E;
    begin
      for L in ERange loop
        if E(L)/=Zero then return False; end if;
      end loop;
      for I in CRange loop
        for J in CRange loop
          if not IsSharp(C(I,J)) then return False; end if;
        end loop;
      end loop;
    end;
    return True;
  end IsSharp;

  procedure Copy(F1: in Spheric; F2: in out Spheric) is
    C1: CArray renames F1.C;
    C2: CArray renames F2.C;
  begin
    for I in CRange loop
      for J in CRange loop Copy(C1(I,J),C2(I,J)); end loop;
    end loop;
    if Not_Trunc then F2.E := F1.E; end if;
  end Copy;

  procedure Swap(F1,F2: in out Spheric) is
    C1: CArray renames F1.C;
    C2: CArray renames F2.C;
  begin
    for I in CRange loop
      for J in CRange loop Swap(C1(I,J),C2(I,J)); end loop;
    end loop;
    if Not_Trunc then
      declare
        E1: constant EArray := F1.E;
      begin
        F1.E := F2.E;
        F2.E := E1;
      end;
    end if;
  end Swap;

  procedure Neg(F: in out Spheric) is
    C: CArray renames F.C;
  begin
    for I in CRange loop
      for J in CRange loop Neg(C(I,J)); end loop;
    end loop;
  end Neg;

  procedure Neg(F1: in Spheric; F2: in out Spheric) is
    C1: CArray renames F1.C;
    C2: CArray renames F2.C;
  begin
    for I in CRange loop
      for J in CRange loop Neg(C1(I,J),C2(I,J)); end loop;
    end loop;
    if Not_Trunc then F2.E := F1.E; end if;
  end Neg;

  function "-"(F: Spheric) return Spheric is
    G: Spheric;
  begin
    Neg(F,G);
    return G;
  end "-";

  procedure Add(S: in SCoeff; F: in out Spheric) is
  begin
    Add(S,F.C(0,0));
  end Add;

  procedure Add(F1: in Spheric; F2: in out Spheric) is
    C1: CArray renames F1.C;
    C2: CArray renames F2.C;
  begin
    for I in CRange loop
      for J in CRange loop Add(C1(I,J),C2(I,J)); end loop;
    end loop;
    if Not_Trunc then
      declare
        E1: EArray renames F1.E;
        E2: EArray renames F2.E;
      begin
        for L in ERange loop E2(L) := E2(L)+E1(L); end loop;
      end;
    end if;
  end Add;

  procedure Sum(F1,F2: in Spheric; F3: in out Spheric) is
    C1: CArray renames F1.C;
    C2: CArray renames F2.C;
    C3: CArray renames F3.C;
  begin
    for I in CRange loop
      for J in CRange loop Sum(C1(I,J),C2(I,J),C3(I,J)); end loop;
    end loop;
    if Not_Trunc then
      declare
        E1: EArray renames F1.E;
        E2: EArray renames F2.E;
        E3: EArray renames F3.E;
      begin
        for L in ERange loop E3(L) := E2(L)+E1(L); end loop;
      end;
    end if;
  end Sum;

  function "+"(F1,F2: Spheric) return Spheric is
    F3: Spheric;
  begin
    Sum(F1,F2,F3);
    return F3;
  end "+";

  procedure Sub(F1: in Spheric; F2: in out Spheric) is
    C1: CArray renames F1.C;
    C2: CArray renames F2.C;
  begin
    for I in CRange loop
      for J in CRange loop Sub(C1(I,J),C2(I,J)); end loop;
    end loop;
    if Not_Trunc then
      declare
        E1: EArray renames F1.E;
        E2: EArray renames F2.E;
      begin
        for L in ERange loop E2(L) := E2(L)+E1(L); end loop;
      end;
    end if;
  end Sub;

  procedure Diff(F1,F2: in Spheric; F3: in out Spheric) is
    C1: CArray renames F1.C;
    C2: CArray renames F2.C;
    C3: CArray renames F3.C;
  begin
    for I in CRange loop
      for J in CRange loop Diff(C1(I,J),C2(I,J),C3(I,J)); end loop;
    end loop;
    if Not_Trunc then
      declare
        E1: EArray renames F1.E;
        E2: EArray renames F2.E;
        E3: EArray renames F3.E;
      begin
        for L in ERange loop E3(L) := E2(L)+E1(L); end loop;
      end;
    end if;
  end Diff;

  function "-"(F1,F2: Spheric) return Spheric is
    F3: Spheric;
  begin
    Diff(F1,F2,F3);
    return F3;
  end "-";

  procedure Mult(S: in Scalar; F: in out Spheric) is
  begin
    if IsZero(S) then
      FillZero(F);
    else
      declare
        C: CArray renames F.C;
      begin
        for I in CRange loop
          for J in CRange loop Mult(S,C(I,J)); end loop;
        end loop;
      end;
      if Not_Trunc then
        declare
          E: EArray renames F.E;
          R: constant Radius := MaxNorm(S);
        begin
          for L in ERange loop E(L) := R*E(L); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Mult(S: in SCoeff; F: in out Spheric) is
  begin
    if IsZero(S) then
      FillZero(F);
    else
      declare
        C: CArray renames F.C;
      begin
        for I in CRange loop
          for J in CRange loop Mult(S,C(I,J)); end loop;
        end loop;
      end;
      if Not_Trunc then
        declare
          E: EArray renames F.E;
          R: constant Radius := MaxNorm(S);
        begin
          for L in ERange loop E(L) := R*E(L); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Prod(S: in Scalar; F1: in Spheric; F2: in out Spheric) is
    C1: CArray renames F1.C;
    C2: CArray renames F2.C;
  begin
    for I in CRange loop
      for J in CRange loop Prod(S,C1(I,J),C2(I,J)); end loop;
    end loop;
    if Not_Trunc then
      declare
        E1: EArray renames F1.E;
        E2: EArray renames F2.E;
        R: constant Radius := MaxNorm(S);
      begin
        if R=Zero then
          E2 := (others => Zero);
        else
          for L in ERange loop E2(L) := R*E1(L); end loop;
        end if;
      end;
    end if;
  end Prod;

  procedure Prod(S: in SCoeff; F1: in Spheric; F2: in out Spheric) is
    SP: Parity renames S.P;
    C1: CArray renames F1.C;
    C2: CArray renames F2.C;
  begin
    if IsZero(S) then
      for I in CRange loop
        for J in CRange loop SetZero((SP+C1(I,J).P) mod 2,C2(I,J)); end loop;
      end loop;
      if Not_Trunc then F2.E := (others => Zero); end if;
    else
      for I in CRange loop
        for J in CRange loop Prod(S,C1(I,J),C2(I,J)); end loop;
      end loop;
      if Not_Trunc then
        declare
          E1: EArray renames F1.E;
          E2: EArray renames F2.E;
          R: constant Radius := MaxNorm(S);
        begin
          for L in ERange loop E2(L) := R*E1(L); end loop;
        end;
      end if;
    end if;
  end Prod;

  function "*"(S: Scalar; F: Spheric) return Spheric is
    G: Spheric;
  begin
    Prod(S,F,G);
    return G;
  end "*";

  function "*"(S: SCoeff; F: Spheric) return Spheric is
    G: Spheric;
  begin
    Prod(S,F,G);
    return G;
  end "*";

  procedure AddProd(S: in Scalar; F1: in Spheric; F2: in out Spheric) is
  begin
    if not IsZero(S) then
      declare
        C1: CArray renames F1.C;
        C2: CArray renames F2.C;
      begin
        for I in CRange loop
          for J in CRange loop AddProd(S,C1(I,J),C2(I,J)); end loop;
        end loop;
      end;
      if Not_Trunc then
        declare
          E1: EArray renames F1.E;
          E2: EArray renames F2.E;
          R: constant Radius := MaxNorm(S);
        begin
          for L in ERange loop E2(L) := E2(L)+R*E1(L); end loop;
        end;
      end if;
    end if;
  end AddProd;

  procedure AddProd(S: in SCoeff; F1: in Spheric; F2: in out Spheric) is
  begin
    if not IsZero(S) then
      declare
        C1: CArray renames F1.C;
        C2: CArray renames F2.C;
      begin
        for I in CRange loop
          for J in CRange loop AddProd(S,C1(I,J),C2(I,J)); end loop;
        end loop;
      end;
      if Not_Trunc then
        declare
          E1: EArray renames F1.E;
          E2: EArray renames F2.E;
          R: constant Radius := MaxNorm(S);
        begin
          for L in ERange loop E2(L) := E2(L)+R*E1(L); end loop;
        end;
      end if;
    end if;
  end AddProd;

  procedure Mult(F1: in Spheric; F2: in out Spheric) is
    F3: Spheric;
  begin
    Prod(F1,F2,F3);
    Swap(F3,F2);
  end Mult;

  ---------------------------------
  --- begin product of Sherical ---
  ---------------------------------

  procedure MaxNorms(C: in CArray; A: out FArray) is
  begin
    for I in CRange loop
      for J in CRange loop A(I,J) := MaxNorm(C(I,J)); end loop;
    end loop;
  end MaxNorms;

  procedure Prod(F1,F2: in Spheric; P3: in Parity; F3: in out Spheric) is
    --- P3 is known parity of F3
    --- uses AddProdC and AddProdE defined below
    C1: CArray renames F1.C;
    C2: CArray renames F2.C;
    E1: EArray renames F1.E;
    E2: EArray renames F2.E;
    E3: EArray renames F3.E;
    ZC1: constant Boolean := IsZero(C1);
    ZC2: constant Boolean := IsZero(C2);
    ZE1: constant Boolean := IsZero(E1);
    ZE2: constant Boolean := IsZero(E2);
    A1,A2: FArray;
  begin
    SetZero(P3,Mixed,F3);
    if ZC1 then
      if ZE1 then return; end if;
      A1 := (others => (others => Zero));
    else
      MaxNorms(C1,A1);
    end if;
    if ZC2 then
      if ZE2 then return; end if;
      A2 := (others => (others => Zero));
    else
      MaxNorms(C2,A2);
    end if;
    if not (ZC1 or else ZC2) then
      Initialize_Regge(PProd_Parallel);
      AddProdC(C1,C2,A1,A2,F3);
      Mult(SHalf,F3);
    end if;
    if not (ZC1 or else ZE2) then
      AddProdE(A1,E2,E3);
    end if;
    if not (ZC2 or else ZE1) then
      AddProdE(A2,E1,E3);
    end if;
    if not (ZE1 or else ZE2) then
      E3(0) := E3(0)+Norm(E1)*Norm(E2);
    end if;
    Sanitize_Parities(F3);
  end Prod;

  procedure Prod(F1,F2: in Spheric; F3: in out Spheric) is
    --- uses AddProdC and AddProdE defined below
    C1: CArray renames F1.C;
    C2: CArray renames F2.C;
    E1: EArray renames F1.E;
    E2: EArray renames F2.E;
    E3: EArray renames F3.E;
    ZC1: constant Boolean := IsZero(C1);
    ZC2: constant Boolean := IsZero(C2);
    ZE1: constant Boolean := IsZero(E1);
    ZE2: constant Boolean := IsZero(E2);
    Is_Zero: Boolean;
    P1,P2: Parity;
    A1,A2: FArray;
  begin
    if ZC1 then
      if ZE1 then SetZero(F3); return; end if; -- cannot guess parity of F3
      A1 := (others => (others => Zero));
      P1 := Par;
    else
      Guess_Parity(F1,Mixed,P1,Is_Zero);
      MaxNorms(C1,A1);
    end if;
    if ZC2 then
      if ZE2 then SetZero(F3); return; end if; -- cannot guess parity of F3
      A2 := (others => (others => Zero));
      P2 := Par;
    else
      Guess_Parity(F2,Mixed,P2,Is_Zero);
      MaxNorms(C2,A2);
    end if;
    SetZero((P1+P2) mod 2,Mixed,F3);           -- guess can improve speed
    if not (ZC1 or else ZC2) then
      Initialize_Regge(PProd_Parallel);
      AddProdC(C1,C2,A1,A2,F3);
      Mult(SHalf,F3);
    end if;
    if not (ZC1 or else ZE2) then
      AddProdE(A1,E2,E3);
    end if;
    if not (ZC2 or else ZE1) then
      AddProdE(A2,E1,E3);
    end if;
    if not (ZE1 or else ZE2) then
      E3(0) := E3(0)+Norm(E1)*Norm(E2);
    end if;
    Sanitize_Parities(F3);
  end Prod;

  procedure AddNorms(B: in BigEArray; E: in out EArray) is
    --- add norms of B to E
    --- use only if Not_Trunc
    R: Flt;
  begin
    E(0) := B(0,0);
    for L in 1 .. SizeE loop
      R := B(L,L);
      for K in 0 .. L-1 loop
        R := R+B(L,K)+B(K,L);
      end loop;
      E(L) := E(L)+Rho2Power(L)*R;
    end loop;
  end AddNorms;

  procedure AddProdC(C1,C2: in CArray; A1,A2: in FArray; F3: in out Spheric) is
    --- uses AddProdC defined below
    --- result needs to be multiplied by 1/2 at the end
    C3: CArray renames F3.C;
    Task_Error: Boolean := False;
    PC: Protected_Counter(0);
    B3: BigEArray := (others => (others => Zero));

    procedure Prod_Loop is
      M3: Natural;
    begin
      Proper_Rounding;
      loop
        PC.Next(M3);
        exit when M3>LastL3;
        AddProdC(M3,C1,C2,A1,A2,C3,B3);
      end loop;
    end Prod_Loop;

    NT: constant Natural := Reserve_Tasks(LastL3,PProd_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is begin
          Prod_Loop;
        exception
          when others => Task_Error := True; raise;
        end Prod_Task_Type;

        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
    if Not_Trunc then AddNorms(B3,F3.E); end if;
  end AddProdC;

  procedure Double(M3: in ERange; C3: in out CArray; B3: in out BigEArray) is
    --- double the result for fixed M3
    K3: Integer;
  begin
    if M3=0 then
      for L3 in 0 .. Size loop
        Mult(Two,C3(L3,L3));
      end loop;
    else
      for L3 in M3 .. Size loop
        K3 := L3-M3;
        Mult(Two,C3(L3,K3));
        Mult(Two,C3(K3,L3));
      end loop;
    end if;
    if Not_Trunc then
      if M3=0 then
        for L3 in 0 .. SizeE loop
          B3(L3,L3) := Two*B3(L3,L3);
        end loop;
      else
        for L3 in M3 .. SizeE loop
          K3 := L3-M3;
          B3(L3,K3) := Two*B3(L3,K3);
          B3(K3,L3) := Two*B3(K3,L3);
        end loop;
      end if;
    end if;
  end Double;

  procedure AddProdC(M3: in ERange; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray) is
    --- add C1*C2 to (C3,B3) with M3 fixed
    --- result needs to be multiplied by 1/2 at the end
    --- uses AddProd_Sum and AddProd_Diff defined below
    --- summing over M1,M2 in 0 .. Size   but restrict to M1+M2=M3 or |M1-M2|=M3
    M1: Integer;
  begin
    --- M2=0 or M1=0
    if M3=0 then                                   --- M3=0
      AddProd_Sum(0,0,0,C1,C2,A1,A2,C3,B3);        ---      M=(0,0,0)
    else                                           --- M3>0
      AddProd_Sum(M3,0,M3,C1,C2,A1,A2,C3,B3);      ---      M=(M3,0,M3)
      AddProd_Sum(M3,0,M3,C2,C1,A2,A1,C3,B3);      ---      M=(0,M3,M3)
    end if;
    Double(M3,C3,B3);                              --- include AddProd_Diff

    if M3=0 then                                   --- M3=0 but M1>0 and M2>0
      for M2 in 1 .. Size loop
        M1 := M2;
        AddProd_Diff(M1,M2,0,C1,C2,A1,A2,C3,B3);   --- Diff M1-M2=M3=0
      end loop;

    else                                           --- M1>0 M2>0 M3>0
      for M2 in 1 .. M3-1 loop                     --- Sum  M3=M1+M2
        M1 := M3-M2;
        if M1<M2 then
          AddProd_Sum(M2,M1,M3,C2,C1,A2,A1,C3,B3); ---         M2 >  M1 > 0
        else
          AddProd_Sum(M1,M2,M3,C1,C2,A1,A2,C3,B3); ---         M1 >= M2 > 0
        end if;
      end loop;

      for M2 in M3+1 .. Size loop                  --- Diff M3=M2-M1
        M1 := M2-M3;
        AddProd_Diff(M2,M1,M3,C2,C1,A2,A1,C3,B3);  ---         M2 >  M1 > 0
      end loop;

      for M2 in 1 .. Size-M3 loop                  --- Diff M3=M1-M2
        M1 := M2+M3;
        AddProd_Diff(M1,M2,M3,C1,C2,A1,A2,C3,B3);  ---         M1 >  M2 > 0
      end loop;
    end if;
  end AddProdC;

  procedure AddProd_Sum(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray) is
    --- assuming M1+M2=M3 and M1>=M2
    --- result needs to be multiplied by 1/2 at the end
  begin
    AddProd_Sum_CosCos(M1,M2,M3,C1,C2,A1,A2,C3,B3);
    if M2>0 and then M3>0 then
      AddProd_Sum_CosSin(M1,M2,M3,C1,C2,A1,A2,C3,B3);
    end if;
    if M1>0 then
      if M3>0 then
        AddProd_Sum_SinCos(M1,M2,M3,C1,C2,A1,A2,C3,B3);
      end if;
      if M2>0 then
        AddProd_Sum_SinSin(M1,M2,M3,C1,C2,A1,A2,C3,B3);
      end if;
    end if;
  end AddProd_Sum;

  procedure AddProd_Diff(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray) is
    --- assuming M1-M2=M3 and M1>=M2
  begin
    AddProd_Diff_CosCos(M1,M2,M3,C1,C2,A1,A2,C3,B3);
    if M2>0 and then M3>0 then
      AddProd_Diff_CosSin(M1,M2,M3,C1,C2,A1,A2,C3,B3);
    end if;
    if M1>0 then
      if M3>0 then
        AddProd_Diff_SinCos(M1,M2,M3,C1,C2,A1,A2,C3,B3);
      end if;
      if M2>0 then
        AddProd_Diff_SinSin(M1,M2,M3,C1,C2,A1,A2,C3,B3);
      end if;
    end if;
  end AddProd_Diff;

  procedure AddProd_Sum_CosCos(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray) is
    --- assuming M1+M2=M3
    --- factor (-1)^m3 from CG coeffs
    --- result needs to be multiplied by 1/2 at the end
    Fac1: constant Boolean := (M3 mod 2)=0;
    ComputeB12,ComputeC12: Boolean;
    K1,K2,K3,L3,MaxL3: Integer;
    B12: Flt;
    Coe: Scalar;
    C12: SCoeff;
  begin
    for L1 in M1 .. Size loop
      K1 := L1-M1;
      if A1(L1,K1)/=Zero then --- cos
        for L2 in M2 .. Size loop
          K2 := L2-M2;
          if A2(L2,K2)/=Zero then --- cos
            ComputeB12 := True;
            ComputeC12 := True;
            L3 := abs(L1-L2);  --- correct parity
            K3 := L3-M3;       --- correct parity, but want K3>=0
            if K3<0 then
              K3 := K3 mod 2;  --- increase K3 without changing parity
              L3 := M3+K3;
            end if;
            MaxL3 := IMin(LastL3,L1+L2);
            while L3 <= MaxL3 loop
              ProdRegge(L1,L2,L3,M1,M2,-M3,Coe); --- M1+M2-M3=0
              if not IsZero(Coe) then
                if Fac1 then
                  Mult(Flt(2*L3+1),Coe);
                else
                  Mult(Flt((-2)*L3-1),Coe);
                end if;
                K3 := L3-M3;
                if L3<=Size then     --- always True if Trunc
                  if ComputeC12 then
                    Prod(C1(L1,K1),C2(L2,K2),C12);
                    ComputeC12 := False;
                  end if;
                  AddProd(Coe,C12,C3(L3,K3)); --- cos
                else
                  if ComputeB12 then
                    B12 := A1(L1,K1)*A2(L2,K2);
                    ComputeB12 := False;
                  end if;
                  B3(L3,K3) := B3(L3,K3)+MaxNorm(Coe)*B12; -- cos
                end if;
              end if;
              L3 := L3+2;
            end loop; -- over L3
          end if;
        end loop; -- over L2
      end if;
    end loop; -- over L1
  end AddProd_Sum_CosCos;

  procedure AddProd_Diff_CosCos(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray) is
    --- assuming M1-M2=M3
    --- factor (-1)^m3 from CG coeffs
    --- factor (-1)^m2 from Diff
    --- result needs to be multiplied by 1/2 at the end
    Fac1: constant Boolean := (M1 mod 2)=0;
    ComputeB12,ComputeC12: Boolean;
    K1,K2,K3,L3,MaxL3: Integer;
    B12: Flt;
    Coe: Scalar;
    C12: SCoeff;
  begin
    for L1 in M1 .. Size loop
      K1 := L1-M1;
      if A1(L1,K1)/=Zero then --- cos
        for L2 in M2 .. Size loop
          K2 := L2-M2;
          if A2(L2,K2)/=Zero then --- cos
            ComputeB12 := True;
            ComputeC12 := True;
            L3 := abs(L1-L2);  --- correct parity
            K3 := L3-M3;       --- correct parity, but want K3>=0
            if K3<0 then
              K3 := K3 mod 2;  --- increase K3 without changing parity
              L3 := M3+K3;
            end if;
            MaxL3 := IMin(LastL3,L1+L2);
            while L3 <= MaxL3 loop
              ProdRegge(L1,L2,L3,M1,-M2,-M3,Coe); --- M1-M2-M3=0
              if not IsZero(Coe) then
                if Fac1 then
                  Mult(Flt(2*L3+1),Coe);
                else
                  Mult(Flt((-2)*L3-1),Coe);
                end if;
                K3 := L3-M3;
                if L3<=Size then     --- always True if Trunc
                  if ComputeC12 then
                    Prod(C1(L1,K1),C2(L2,K2),C12);
                    ComputeC12 := False;
                  end if;
                  AddProd(Coe,C12,C3(L3,K3)); --- cos
                else
                  if ComputeB12 then
                    B12 := A1(L1,K1)*A2(L2,K2);
                    ComputeB12 := False;
                  end if;
                  B3(L3,K3) := B3(L3,K3)+MaxNorm(Coe)*B12; --- cos
                end if;
              end if;
              L3 := L3+2;
            end loop; -- over L3
          end if;
        end loop; -- over L2
      end if;
    end loop; -- over L1
  end AddProd_Diff_CosCos;

  procedure AddProd_Sum_CosSin(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray) is
    --- assuming M1+M2=M3
    --- factor (-1)^m3 from CG coeffs
    --- result needs to be multiplied by 1/2 at the end
    Fac1: constant Boolean := (M3 mod 2)=0;
    ComputeB12,ComputeC12: Boolean;
    K1,K2,K3,L3,MaxL3: Integer;
    B12: Flt;
    Coe: Scalar;
    C12: SCoeff;
  begin
    for L1 in M1 .. Size loop
      K1 := L1-M1;
      if A1(L1,K1)/=Zero then --- cos
        for L2 in M2 .. Size loop
          K2 := L2-M2;
          if A2(K2,L2)/=Zero then --- sin
            ComputeB12 := True;
            ComputeC12 := True;
            L3 := abs(L1-L2);  --- correct parity
            K3 := L3-M3;       --- correct parity, but want K3>=0
            if K3<0 then
              K3 := K3 mod 2;  --- increase K3 without changing parity
              L3 := M3+K3;
            end if;
            MaxL3 := IMin(LastL3,L1+L2);
            while L3 <= MaxL3 loop
              ProdRegge(L1,L2,L3,M1,M2,-M3,Coe); --- M1+M2-M3=0
              if not IsZero(Coe) then
                if Fac1 then
                  Mult(Flt(2*L3+1),Coe);
                else
                  Mult(Flt((-2)*L3-1),Coe);
                end if;
                K3 := L3-M3;
                if L3<=Size then     --- always True if Trunc
                  if ComputeC12 then
                    Prod(C1(L1,K1),C2(K2,L2),C12);
                    ComputeC12 := False;
                  end if;
                  AddProd(Coe,C12,C3(K3,L3)); --- sin
                else
                  if ComputeB12 then
                    B12 := A1(L1,K1)*A2(K2,L2);
                    ComputeB12 := False;
                  end if;
                  B3(K3,L3) := B3(K3,L3)+MaxNorm(Coe)*B12; --- sin
                end if;
              end if;
              L3 := L3+2;
            end loop; -- over L3
          end if;
        end loop; -- over L2
      end if;
    end loop; -- over L1
  end AddProd_Sum_CosSin;

  procedure AddProd_Diff_CosSin(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray) is
    --- assuming M1-M2=M3
    --- factor (-1)^m3 from CG coeffs
    --- factor -(-1)^m2 from Diff
    --- result needs to be multiplied by 1/2 at the end
    Fac1: constant Boolean := (M1 mod 2)=1;
    ComputeB12,ComputeC12: Boolean;
    K1,K2,K3,L3,MaxL3: Integer;
    B12: Flt;
    Coe: Scalar;
    C12: SCoeff;
  begin
    for L1 in M1 .. Size loop
      K1 := L1-M1;
      if A1(L1,K1)/=Zero then --- cos
        for L2 in M2 .. Size loop
          K2 := L2-M2;
          if A2(K2,L2)/=Zero then --- sin
            ComputeB12 := True;
            ComputeC12 := True;
            L3 := abs(L1-L2);  --- correct parity
            K3 := L3-M3;       --- correct parity, but want K3>=0
            if K3<0 then
              K3 := K3 mod 2;  --- increase K3 without changing parity
              L3 := M3+K3;
            end if;
            MaxL3 := IMin(LastL3,L1+L2);
            while L3 <= MaxL3 loop
              ProdRegge(L1,L2,L3,M1,-M2,-M3,Coe); --- M1-M2-M3=0
              if not IsZero(Coe) then
                if Fac1 then
                  Mult(Flt(2*L3+1),Coe);
                else
                  Mult(Flt((-2)*L3-1),Coe);
                end if;
                K3 := L3-M3;
                if L3<=Size then     --- always True if Trunc
                  if ComputeC12 then
                    Prod(C1(L1,K1),C2(K2,L2),C12);
                    ComputeC12 := False;
                  end if;
                  AddProd(Coe,C12,C3(K3,L3)); --- sin
                else
                  if ComputeB12 then
                    B12 := A1(L1,K1)*A2(K2,L2);
                    ComputeB12 := False;
                  end if;
                  B3(K3,L3) := B3(K3,L3)+MaxNorm(Coe)*B12; --- sin
                end if;
              end if;
              L3 := L3+2;
            end loop; -- over L3
          end if;
        end loop; -- over L2
      end if;
    end loop; -- over L1
  end AddProd_Diff_CosSin;

  procedure AddProd_Sum_SinCos(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray) is
    --- assuming M1+M2=M3
    --- factor (-1)^m3 from CG coeffs
    --- result needs to be multiplied by 1/2 at the end
    Fac1: constant Boolean := (M3 mod 2)=0;
    ComputeB12,ComputeC12: Boolean;
    K1,K2,K3,L3,MaxL3: Integer;
    B12: Flt;
    Coe: Scalar;
    C12: SCoeff;
  begin
    for L1 in M1 .. Size loop
      K1 := L1-M1;
      if A1(K1,L1)/=Zero then --- sin
        for L2 in M2 .. Size loop
          K2 := L2-M2;
          if A2(L2,K2)/=Zero then --- cos
            ComputeB12 := True;
            ComputeC12 := True;
            L3 := abs(L1-L2);  --- correct parity
            K3 := L3-M3;       --- correct parity, but want K3>=0
            if K3<0 then
              K3 := K3 mod 2;  --- increase K3 without changing parity
              L3 := M3+K3;
            end if;
            MaxL3 := IMin(LastL3,L1+L2);
            while L3 <= MaxL3 loop
              ProdRegge(L1,L2,L3,M1,M2,-M3,Coe); --- M1+M2-M3=0
              if not IsZero(Coe) then
                if Fac1 then
                  Mult(Flt(2*L3+1),Coe);
                else
                  Mult(Flt((-2)*L3-1),Coe);
                end if;
                K3 := L3-M3;
                if L3<=Size then     --- always True if Trunc
                  if ComputeC12 then
                    Prod(C1(K1,L1),C2(L2,K2),C12);
                    ComputeC12 := False;
                  end if;
                  AddProd(Coe,C12,C3(K3,L3)); --- sin
                else
                  if ComputeB12 then
                    B12 := A1(K1,L1)*A2(L2,K2);
                    ComputeB12 := False;
                  end if;
                  B3(K3,L3) := B3(K3,L3)+MaxNorm(Coe)*B12; -- sin
                end if;
              end if;
              L3 := L3+2;
            end loop; -- over L3
          end if;
        end loop; -- over L2
      end if;
    end loop; -- over L1
  end AddProd_Sum_SinCos;

  procedure AddProd_Diff_SinCos(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray) is
    --- assuming M1-M2=M3
    --- factor (-1)^m3 from CG coeffs
    --- factor (-1)^m2 from Diff
    --- result needs to be multiplied by 1/2 at the end
    Fac1: constant Boolean := (M1 mod 2)=0;
    ComputeB12,ComputeC12: Boolean;
    K1,K2,K3,L3,MaxL3: Integer;
    B12: Flt;
    Coe: Scalar;
    C12: SCoeff;
  begin
    for L1 in M1 .. Size loop
      K1 := L1-M1;
      if A1(K1,L1)/=Zero then --- sin
        for L2 in M2 .. Size loop
          K2 := L2-M2;
          if A2(L2,K2)/=Zero then --- cos
            ComputeB12 := True;
            ComputeC12 := True;
            L3 := abs(L1-L2);  --- correct parity
            K3 := L3-M3;       --- correct parity, but want K3>=0
            if K3<0 then
              K3 := K3 mod 2;  --- increase K3 without changing parity
              L3 := M3+K3;
            end if;
            MaxL3 := IMin(LastL3,L1+L2);
            while L3 <= MaxL3 loop
              ProdRegge(L1,L2,L3,M1,-M2,-M3,Coe); --- M1-M2-M3=0
              if not IsZero(Coe) then
                if Fac1 then
                  Mult(Flt(2*L3+1),Coe);
                else
                  Mult(Flt((-2)*L3-1),Coe);
                end if;
                K3 := L3-M3;
                if L3<=Size then     --- always True if Trunc
                  if ComputeC12 then
                    Prod(C1(K1,L1),C2(L2,K2),C12);
                    ComputeC12 := False;
                  end if;
                  AddProd(Coe,C12,C3(K3,L3)); --- sin
                else
                  if ComputeB12 then
                    B12 := A1(K1,L1)*A2(L2,K2);
                    ComputeB12 := False;
                  end if;
                  B3(K3,L3) := B3(K3,L3)+MaxNorm(Coe)*B12; --- sin
                end if;
              end if;
              L3 := L3+2;
            end loop; -- over L3
          end if;
        end loop; -- over L2
      end if;
    end loop; -- over L1
  end AddProd_Diff_SinCos;

  procedure AddProd_Sum_SinSin(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray) is
    --- assuming M1+M2=M3
    --- factor (-1)^m3 from CG coeffs
    --- factor -1 from Sum
    --- result needs to be multiplied by 1/2 at the end
    Fac1: constant Boolean := (M3 mod 2)=1;
    ComputeB12,ComputeC12: Boolean;
    K1,K2,K3,L3,MaxL3: Integer;
    B12: Flt;
    Coe: Scalar;
    C12: SCoeff;
  begin
    for L1 in M1 .. Size loop
      K1 := L1-M1;
      if A1(K1,L1)/=Zero then --- sin
        for L2 in M2 .. Size loop
          K2 := L2-M2;
          if A2(K2,L2)/=Zero then --- sin
            ComputeB12 := True;
            ComputeC12 := True;
            L3 := abs(L1-L2);  --- correct parity
            K3 := L3-M3;       --- correct parity, but want K3>=0
            if K3<0 then
              K3 := K3 mod 2;  --- increase K3 without changing parity
              L3 := M3+K3;
            end if;
            MaxL3 := IMin(LastL3,L1+L2);
            while L3 <= MaxL3 loop
              ProdRegge(L1,L2,L3,M1,M2,-M3,Coe); --- M1+M2-M3=0
              if not IsZero(Coe) then
                if Fac1 then
                  Mult(Flt(2*L3+1),Coe);
                else
                  Mult(Flt((-2)*L3-1),Coe);
                end if;
                K3 := L3-M3;
                if L3<=Size then     --- always True if Trunc
                  if ComputeC12 then
                    Prod(C1(K1,L1),C2(K2,L2),C12);
                    ComputeC12 := False;
                  end if;
                  AddProd(Coe,C12,C3(L3,K3)); --- cos
                else
                  if ComputeB12 then
                    B12 := A1(K1,L1)*A2(K2,L2);
                    ComputeB12 := False;
                  end if;
                  B3(L3,K3) := B3(L3,K3)+MaxNorm(Coe)*B12; -- cos
                end if;
              end if;
              L3 := L3+2;
            end loop; -- over L3
          end if;
        end loop; -- over L2
      end if;
    end loop; -- over L1
  end AddProd_Sum_SinSin;

  procedure AddProd_Diff_SinSin(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray) is
    --- assuming M1-M2=M3
    --- factor (-1)^m3 from CG coeffs
    --- factor (-1)^m2 from Diff
    --- result needs to be multiplied by 1/2 at the end
    Fac1: constant Boolean := (M1 mod 2)=0;
    ComputeB12,ComputeC12: Boolean;
    K1,K2,K3,L3,MaxL3: Integer;
    B12: Flt;
    Coe: Scalar;
    C12: SCoeff;
  begin
    for L1 in M1 .. Size loop
      K1 := L1-M1;
      if A1(K1,L1)/=Zero then --- sin
        for L2 in M2 .. Size loop
          K2 := L2-M2;
          if A2(K2,L2)/=Zero then --- sin
            ComputeB12 := True;
            ComputeC12 := True;
            L3 := abs(L1-L2);  --- correct parity
            K3 := L3-M3;       --- correct parity, but want K3>=0
            if K3<0 then
              K3 := K3 mod 2;  --- increase K3 without changing parity
              L3 := M3+K3;
            end if;
            MaxL3 := IMin(LastL3,L1+L2);
            while L3 <= MaxL3 loop
              ProdRegge(L1,L2,L3,M1,-M2,-M3,Coe); --- M1-M2-M3=0
              if not IsZero(Coe) then
                if Fac1 then
                  Mult(Flt(2*L3+1),Coe);
                else
                  Mult(Flt((-2)*L3-1),Coe);
                end if;
                K3 := L3-M3;
                if L3<=Size then     --- always True if Trunc
                  if ComputeC12 then
                    Prod(C1(K1,L1),C2(K2,L2),C12);
                    ComputeC12 := False;
                  end if;
                  AddProd(Coe,C12,C3(L3,K3)); --- cos
                else
                  if ComputeB12 then
                    B12 := A1(K1,L1)*A2(K2,L2);
                    ComputeB12 := False;
                  end if;
                  B3(L3,K3) := B3(L3,K3)+MaxNorm(Coe)*B12; --- cos
                end if;
              end if;
              L3 := L3+2;
            end loop; -- over L3
          end if;
        end loop; -- over L2
      end if;
    end loop; -- over L1
  end AddProd_Diff_SinSin;

  procedure AddProdE(A1: in FArray; E2: in EArray; E3: in out EArray) is
    L3: Integer;
    E30: Radius renames E3(0);
    EL1: Radius;
  begin
    for L1 in CRange loop
      EL1 := A1(L1,L1);
      for K1 in 0 .. L1-1 loop
        EL1 := EL1+A1(L1,K1)+A1(K1,L1);
      end loop;
      if EL1/=Zero then
        EL1 := Rho2Power(L1)*EL1; --- analogue of E1(L)
        for L2 in ERange loop
          if E2(L2)/=Zero then
            L3 := L2-L1;
            if L3>0 then
              E3(L3) := E3(L3)+EL1*E2(L2);
            else
              E30 := E30+EL1*E2(L2);
            end if;
          end if;
        end loop; -- L2
      end if;
    end loop; -- L1
  end AddProdE;

  --------------------------------
  --- end product of Spherical ---
  --------------------------------

  function "*"(F1,F2: Spheric) return Spheric is
    F3: Spheric;
  begin
    Prod(F1,F2,F3);
    return F3;
  end "*";

  function MaxNorm(F: Spheric) return Radius is
  begin
    if Trunc then return MaxNorm(F.C); end if;
    return MaxNorm(F.C)+Norm(F.E);
  end MaxNorm;

  procedure Norm(F: in Spheric; S: in out Scalar) is
    C: CArray renames F.C;
    T,Tmp: Scalar;
  begin
    SetZero(S);
    for L in reverse CRange loop
      Norm(C(L,L),T);            -- M=0
      for K in 0 .. L-1 loop     -- K=L-M
        Norm(C(L,K),Tmp);
        Add(Tmp,T);
        Norm(C(K,L),Tmp);
        Add(Tmp,T);
      end loop;
      Mult(SRho,S);
      Add(T,S);
    end loop;
    if Not_Trunc then
      BallAt0(Norm(F.E),T);
      Add(T,S);
    end if;
  end Norm;

  function Norm(F: Spheric) return Scalar is
    S: Scalar;
  begin
    Norm(F,S);
    return S;
  end Norm;

  procedure Coeff(L: in Natural; M0: in Integer; F: in Spheric; S: in out SCoeff) is
    --- M0>=0 --> cos coeff, M0<0 --> sin coeff
    M: constant Integer := abs(M0);
  begin
    if M>L then raise Undefined; end if;
    if L<=Size then
      if M0<0 then Copy(F.C(L-M,L),S); else Copy(F.C(L,L-M),S); end if;
    end if;
    if Not_Trunc then
      declare
        E: EArray renames F.E;
        R: Radius := E(0);
      begin
        for K in 1 .. L loop
          R := InvRho2*R;
          if K<=SizeE then R := R+E(K); end if;
        end loop;
        AddBall(R,S);
      end;
    end if;
  end Coeff;

  procedure AddCoeff(L: in Natural; M0: in Integer; S: in SCoeff; F: in out Spheric) is
    --- M0>=0 --> cos coeff, M0<0 --> sin coeff
    M: constant Integer := abs(M0);
  begin
    if M>L then raise Undefined; end if;
    if L>Size then
      if Not_Trunc then
        if L>SizeE then
          F.E(SizeE) := F.E(SizeE)+Rho2Power(L)*MaxNorm(S);
        else
          F.E(L) := F.E(L)+Rho2Power(L)*MaxNorm(S);
        end if;
      end if;
    elsif M0<0 then
      Add(S,F.C(L-M,L)); -- sin
    else
      Add(S,F.C(L,L-M)); -- cos
    end if;
  end AddCoeff;

  procedure AddHigher(L: in Natural; E: in Radius; F: in out Spheric) is
  begin
    if Trunc then
      null;
    elsif L>SizeE then
      F.E(SizeE) := F.E(SizeE)+E;
    else
      F.E(L) := F.E(L)+E;
    end if;
  end AddHigher;

  procedure AddBall(R: in Radius; F: in out Spheric) is
  begin
    if Not_Trunc then F.E(0) := F.E(0)+R; end if;
  end AddBall;

  procedure AddBall(S: in Scalar; F: in out Spheric) is
  begin
    if Not_Trunc then F.E(0) := F.E(0)+MaxNorm(S); end if;
  end AddBall;

  procedure Show_Parities(N: in String; F: in Spheric; Hide0: in Boolean) is
    W: constant Positive := Width(Size);
    C: CArray renames F.C;
  begin
    Show0(N & " cos(phi) t-parities");
    for L in 0 .. Size loop
      Show0("L=" & Strng(L,W) & " Par: ",False);
      for K in 0 .. L loop
        if not (Hide0 and then IsZero(C(L,K))) then
          if C(L,K).P=0 then Show0("0",False); else Show0("1",False); end if;
        end if;
      end loop;
      Show0("");
    end loop;
    Show0(N & " sin(phi) t-parities");
    for L in 1 .. Size loop
      Show0("L=" & Strng(L,W) & " Par: ",False);
      for K in 0 .. L-1 loop
        if not (Hide0 and then IsZero(C(K,L))) then
          if C(K,L).P=0 then Show0("0",False); else Show0("1",False); end if;
        end if;
      end loop;
      Show0("");
    end loop;
  end Show_Parities;

  procedure Show_Guessed_Radii(N: in String; F: in Spheric; Cut: in Radius) is
    W: constant Positive := Width(Size);
    C: CArray renames F.C;
    K: Integer;
    R: Radius;
  begin
    if FDeg=0 then
      Show0(N & "independent of time");
    else
      for L in 0 .. Size loop
        for M in 0 .. L loop
          K := L-M;
          if MaxNorm(C(L,K))>Cut then              --- cos(m*phi) coeff
            R := GuessRho(C(L,K));
            Show1(N & "(L,M)=(" & Strng(L,W) & "," & Strng(M,W) & ") phi-even radius =",R);
          end if;
          if M>0 and then MaxNorm(C(K,L))>Cut then --- sin(m*phi) coeff
            R := GuessRho(C(K,L));
            Show1(N & "(L,M)=(" & Strng(L,W) & "," & Strng(M,W) & ") phi-odd  radius =",R);
          end if;
        end loop;
      end loop;
    end if;
  end Show_Guessed_Radii;

  procedure Show(N: in String; F: in Spheric; Cut: in Radius; Hide0: in Boolean := True) is
    NotHide0: constant Boolean := not Hide0;
    W: constant Positive := Width(SizeE);
    K: Integer;
    C: CArray renames F.C;
    function WStr(I: Integer) return String is begin return Strings.Strng(I,W); end WStr;
  begin
    if Cut>Zero then Show1(N & "Cutoff ",Cut); end if;
    if MaxNorm(F)>Cut then
      Show1(N & "---------------------- Rho ",Rho2);
      for L in CRange loop
        for M in 0 .. L loop
          K := L-M;
          Show(N & "Cos: (l,m)=(" & WStr(L) & ", " & WStr(M) & ") C: ",C(L,K),Cut,Hide0);
          if M>0 then
            Show(N & "Sin: (l,m)=(" & WStr(L) & ", " & WStr(M) & ") C: ",C(K,L),Cut,Hide0);
          end if;
        end loop;
      end loop;
      if Not_Trunc then
        declare
          E: EArray renames F.E;
        begin
          for L in ERange loop
            if NotHide0 or else E(L)>Cut then
              Show1(N & "err: (l,m)=(" & WStr(L) & ", " & WStr(0) & ") E: ",E(L));
            end if;
          end loop;
        end;
      end if;
    elsif NotHide0 then
      Show0(N & "0");
    end if;
  end Show;

  procedure Show1(N: in String; F: in Spheric; Hide0: in Boolean := True) is
  begin
    Show(N,F,Zero,Hide0);
  end Show1;

  procedure Put(F: in File_Type; H: in Spheric; Decimal: in Boolean := False) is
    Simple: constant Boolean := FDeg=0;
    C: CArray renames H.C;
    E: EArray renames H.E;
    LC: constant Integer := Imax(0,EffLast(C));
    LE: constant Integer := EffLast(E);
  begin
    Put(F,FDeg);         --- 0 means Simple
    Put(F,Rho1,Decimal);
    Put(F,Rho2,Decimal);
    Put(F,LC);
    for L in 0 .. LC loop
      Put(F,C(L,L),Simple);
      for M in 1 .. L loop
        Put(F,C(L,L-M),Simple); -- cos
        Put(F,C(L-M,L),Simple); -- sin
      end loop;
    end loop;
    if LE<0 then
      Put(F,Integer(0));
      Put(F,Zero,Decimal);
    else
      Put(F,LE);
      for L in 0 .. LE loop Put(F,E(L),Decimal); end loop;
    end if;
  end Put;

  procedure Get(F: in File_Type; H: in out Spheric; Decimal: in Boolean := False) is
    Simple: Boolean;            --- SCoeff saved like Scalar
    OldFDeg,LC,LE: Integer := 0;
    OldRad,OldRho,R: Radius := Zero;
    C: SCoeff;
  begin
    SetZero(H);
    Get(F,OldFDeg,Decimal);
    Simple := (OldFDeg=0);
    Get(F,OldRad,Decimal);
    Get(F,OldRho,Decimal);
    if Not_Trunc then
      if OldRad<Rho1 and then not Simple then
        Show2("TSpherics.Get Warning: OldRad<Rad ",OldRad,Rho1);
      end if;
      if OldRho<Rho2 then
        Show2("TSpherics.Get Warning: OldRho<Rho ",OldRho,Rho2);
      end if;
    end if;
    Get(F,LC);
    for L in 0 .. LC loop
      Get(F,C,Simple);
      AddCoeff(L,0,C,H);
      for M in 1 .. L loop
        Get(F,C,Simple);
        AddCoeff(L, M,C,H); -- cos
        Get(F,C,Simple);
        AddCoeff(L,-M,C,H); -- sin
      end loop;
    end loop;
    Get(F,LE);
    for L in 0 .. LE loop
      Get(F,R,Decimal);
      AddHigher(L,R,H);
    end loop;
  end Get;

  procedure Write(FileName: in String; H: in Spheric; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,H,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; H: in out Spheric; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,H,Decimal);
    Close(F);
  end Read;

  procedure Leading_Coeffs(Str: in String; P: in SCoeff; Cut: in Radius) is
    --- used by Leading_Coeffs below
    CS: constant Character := Choose(P.P=0,'C','S');
    W: constant Positive := Width(FDeg);
    R: Radius;
    S: Scalar;
    function WStr(I: Integer) return String is begin return Strings.Strng(I,W); end WStr;
  begin
    if Errs then
      declare
        E: Errs1 renames P.E;
      begin
        R := Zero;
        for N in reverse FDeg+1 .. EDeg loop R := R+E(N); end loop;
        if R>Cut then raise Not_Implemented; end if; --- increase Cut
      end;
    end if;
    for N in P.P .. FDeg loop
      Coeff(N,P,S);
      R := MaxNorm(S);
      if R>Cut then
        Show1("",S,False);
        Show0("  *  " & Str & CS & "(" & WStr(N) & ")");
      end if;
    end loop;
  end Leading_Coeffs;

  procedure Leading_Coeffs(F: in Spheric; Cut: in Radius) is
    W: constant Positive := Width(Size);
    C: CArray renames F.C;
    K: Integer;
    R: Radius;
    P: SCoeff;
    function WStr(I: Integer) return String is begin return Strings.Strng(I,W); end WStr;
  begin
    if Not_Trunc then
      declare
        E: EArray renames F.E;
      begin
        R := Zero;
        for L in reverse Size+1 .. SizeE loop R := R+E(L); end loop;
        if R>Cut then raise Not_Implemented; end if; --- increase Cut
      end;
    end if;
    for L in 0 .. Size loop
      for M in 0 .. L loop
        K := L-M;
        R := MaxNorm(C(L,K));   --- Cos(m*phi)
        if R>Cut then
          Coeff(L,M,F,P);
          Leading_Coeffs("P(" & WStr(L) & "," & WStr(M) & ") C(" & WStr(M) & ") ",P,Cut);
        end if;
        if M>0 then
          R := MaxNorm(C(K,L)); --- Sin(m*phi)
          if R>Cut then
            Coeff(L,-M,F,P);
            Leading_Coeffs("P(" & WStr(L) & "," & WStr(M) & ") S(" & WStr(M) & ") ",P,Cut);
          end if;
        end if;
      end loop;
    end loop;
  end Leading_Coeffs;

  procedure Val(F: in Spheric; Theta,Phi: in Scalar; S: in out SCoeff) is
    --- assuming Theta and Phi are real
    C: CArray renames F.C;
    Tmp: Scalar;
    Co,Si: SVector(CRange);
    V: SArray;
  begin
    ALegendVals(Theta,V);        --- V(L,M)=P^m_l(cos(theta))
    CSPowers(Phi,Co,Si);         --- Co(M)=cos(M*Phi), Si(M)=sin(M*Phi)
    SetZero(S);
    for L in reverse 1 .. Size loop
      for M in reverse 1 .. L loop
        Prod(V(L,M),Co(M),Tmp);  --- P^m_l(cos(theta))cos(m*phi)
        AddProd(Tmp,C(L,L-M),S);
        Prod(V(L,M),Si(M),Tmp);  --- P^m_l(cos(theta))sin(m*phi)
        AddProd(Tmp,C(L-M,L),S);
      end loop;
      AddProd(V(L,0),C(L,L),S);  --- add m=0 term
    end loop;
    AddProd(V(0,0),C(0,0),S);    --- add l=0 term
    if Not_Trunc then
      AddBall(Norm(F.E),S);
    end if;
  end Val;

  procedure Val(F: in Spheric; T,Theta,Phi: in Scalar; S: in out Scalar) is
    --- T is time
    P: SCoeff;
  begin
    Val(F,Theta,Phi,P);
    Val(P,T,S);
  end Val;

  procedure Val2(F: in Spheric; Theta,Phi: in Scalar; S: in out SCoeff2) is
    --- assuming Theta and Phi are real
    C: CArray renames F.C;
    Tmp: Scalar;
    Co,Si: SVector(CRange);
    V: SArray;
  begin
    if Trunc then
      SetZero(0,S.E);
      SetZero(1,S.O);
    elsif Mixed or Par=0 then
      BallAt0(Norm(F.E),0,S.E);
      SetZero(1,S.O);
    elsif Par=1 then             --- all odd in t
      SetZero(0,S.E);
      BallAt0(Norm(F.E),1,S.E);
    end if;
    ALegendVals(Theta,V);        --- V(L,M)=P^m_l(cos(theta))
    CSPowers(Phi,Co,Si);         --- Co(M)=cos(M*Phi), Si(M)=sin(M*Phi)
    for L in reverse 1 .. Size loop
      for M in reverse 1 .. L loop
        Prod(V(L,M),Co(M),Tmp);  --- P^m_l(cos(theta))cos(m*phi)
        AddProd(Tmp,C(L,L-M),S);
        Prod(V(L,M),Si(M),Tmp);  --- P^m_l(cos(theta))sin(m*phi)
        AddProd(Tmp,C(L-M,L),S);
      end loop;
      AddProd(V(L,0),C(L,L),S);  --- add m=0 term
    end loop;
    AddProd(V(0,0),C(0,0),S);    --- add l=0 term
  end Val2;

  procedure Val2(F: in Spheric; T,Theta,Phi: in Scalar; S: in out Scalar) is
    --- T is time
    P: SCoeff2;
  begin
    Val2(F,Theta,Phi,P);
    Val(P,T,S);
  end Val2;

  procedure NegInvLap(F: in out Spheric) is
    --- only on sphere, no time dependence
    C: CArray renames F.C;
    InvEigen: Scalar;
  begin
    if not IsZero(C(0,0)) then raise Undefined; end if;
    for L in 1 .. Size loop
      Quot(SOne,Flt(L*(L+1)),InvEigen);
      Mult(InvEigen,C(L,L));   -- M=0
      for K in 0 .. L-1 loop   -- K=L-M
        Mult(InvEigen,C(L,K)); -- cos
        Mult(InvEigen,C(K,L)); -- sin
      end loop;
    end loop;
    if Not_Trunc then
      declare
        E: EArray renames F.E;
      begin
        E(1) := E(1)+E(0);
        E(0) := Zero;
        for L in 1 .. SizeE loop
          E(L) := E(L)/Flt(L*(L+1));
        end loop;
      end;
    end if;
  end NegInvLap;

  procedure NegInvLap(F1: in Spheric; F2: in out Spheric) is
  begin
    Copy(F1,F2);
    NegInvLap(F2);
  end NegInvLap;

  --------------------------
  --- periodic colutions ---
  --------------------------

  procedure InvPlate(F: in out Spheric) is
    C: CArray renames F.C;
    Del: Radius;
    LL1: Scalar;
  begin
    if FreqPar>=0 then raise Not_Implemented; end if;
    ProjAll(F);
    if not IsZero(C(0,0)) then raise Undefined; end if;
    for L in 1 .. Size loop
      Plate_Del(L,Del);                  -- Del only needed for largish L
      Assign(L*(L+1),LL1);
      InvHelm(SAlpha,LL1,Del,C(L,L));    -- M=0
      for K in 0 .. L-1 loop             -- K=L-M
        InvHelm(SAlpha,LL1,Del,C(L,K));  -- cos
        InvHelm(SAlpha,LL1,Del,C(K,L));  -- sin
      end loop;
    end loop;
    if Not_Trunc then
      declare
        FAlphaQ: constant Radius := Flt(AlphaQ);
        E: EArray renames F.E;
      begin
        if E(0)/=Zero then raise Undefined; end if;
        for L in 1 .. SizeE loop
          E(L) := E(L)*FAlphaQ/Flt(L*(L+1));
        end loop;
      end;
    end if;
  end InvPlate;

  procedure InvWave(F: in out Spheric; WDel: in EArray) is
    --- somehow have to enforce that N=0 when L=0
    C: CArray renames F.C;
    Del: Radius;
    LL1,B: Scalar;
  begin
    ProjAll(F);
    for L in CRange loop
      Del := WDel(L);             -- Del only needed for largish L
      Assign(L*(L+1),LL1);
      Sqrt(LL1,B);
      InvHelm(SAlpha,B,Del,C(L,L));    -- M=0
      for K in 0 .. L-1 loop           -- K=L-M
        InvHelm(SAlpha,B,Del,C(L,K));  -- cos
        InvHelm(SAlpha,B,Del,C(K,L));  -- sin
      end loop;
    end loop;
    if Not_Trunc then
      declare
        E: EArray renames F.E;
        MinDist: Radius := Radius'Last;
        Dist,InvB: Radius;
        use Flt_EF;
      begin
        if E(0)/=Zero then raise Undefined; end if;
        for L in reverse 1 .. SizeE loop
          Dist := WDel(L);
          if Dist<MinDist then MinDist := Dist; end if;
          if E(L)/=Zero then
            InvB := Sqrt(One/Flt(L*(L+1)));
            E(L) := E(L)*InvB/MinDist;
          end if;
        end loop;
      end;
    end if;
  end InvWave;

  procedure InvWave(F: in out Spheric) is
  begin
    InvWave(F,Wave_Delta); --- global var Wave_Delta becomes constan
  end InvWave;

  procedure RotWave(F: in out Spheric; WDel: in EArray) is
    --- meant for FDeg=0 and Alpha>1
    C: CArray renames F.C;
    Lam0,Lam,Tmp: Scalar;
  begin
    ProjAll(F);
    if not IsZero(C(0,0)) then raise Undefined; end if;
    for L in 1 .. Size loop
      Assign(L*(L+1),Lam0);         --- Lam0=L*(L+1)
      Div(Lam0,C(L,L));
      for M in 1 .. L loop
        Prod(Flt(M*M),SAlpha2,Tmp);
        Diff(Lam0,Tmp,Lam);         --- Lam=L*(L+1)-(M*Alpha)^2
        Inv(Lam,Tmp);
        Mult(Tmp,C(L,L-M));
        Mult(Tmp,C(L-M,L));
      end loop;
    end loop;
    if Not_Trunc then
      declare
        E: EArray renames F.E;
        MinDist: Radius := Radius'Last;
        Dist,InvB: Radius;
        use Flt_EF;
      begin
        if E(0)/=Zero then raise Undefined; end if;
        for L in reverse 1 .. SizeE loop
          Dist := WDel(L);
          if Dist<MinDist then MinDist := Dist; end if;
          if E(L)/=Zero then
            InvB := Sqrt(One/Flt(L*(L+1)));
            E(L) := E(L)*InvB/MinDist;
          end if;
        end loop;
      end;
    end if;
  end RotWave;

  procedure RotWave(F: in out Spheric) is
  begin
    RotWave(F,Wave_Delta); --- global var Wave_Delta becomes const
  end RotWave;

  ---------------------
  --- quasiperiodic ---
  ---------------------

  procedure QP_InvHelm(L,M: in Positive; P: in Parity; QSig,QDel: in Errs1; C,S: in out SCoeff) is
    --- check parities before using
    --- P=0: (t,phi)-->(-t,-phi) symm
    --- P=1: (t,phi)-->(-t,-phi) antisymm
    CC: FPoly1 renames C.C;
    SC: FPoly1 renames S.C;
    B,B2,BetaM,AlphaN,InvLambdaPlus,InvLambdaMinus,Sigmo,Delto,Tmp1,Tmp2: Scalar;
  begin
    Assign(Flt(L*(L+1)),B);
    Prod(B,B,B2);
    Prod(Flt(M),SBeta,BetaM);
    for N in FFreq loop
      declare
        CN: Scalar renames CC(N);
        SN: Scalar renames SC(N);
      begin
        if not (IsZero(CN) and then IsZero(SN)) then
          Assign(N*Alpha,AlphaN);
          Diff(AlphaN,BetaM,Tmp1);  --- Alpha*N-Beta*M
          Prod(Tmp1,Tmp1,Tmp2);     --- (Alpha*N-Beta*M)^2
          Diff(B2,Tmp2,Tmp1);       --- LambdaMinus = [l(l+1)]^2-(Alpha*N-Beta*M)^2
          Inv(Tmp1,InvLambdaMinus); --- 1/LambdaMinus

          Sum(AlphaN,BetaM,Tmp1);   --- Alpha*N+Beta*M
          Prod(Tmp1,Tmp1,Tmp2);     --- (Alpha*N+Beta*M)^2
          Diff(B2,Tmp2,Tmp1);       --- LambdaPlus = [l(l+1)]^2-(Alpha*N+Beta*M)^2
          Inv(Tmp1,InvLambdaPlus);  --- 1/LambdaPlus

          Sum(InvLambdaMinus,InvLambdaPlus,Sigmo);
          Mult(SHalf,Sigmo);                        --- Sigmo = [1/LambdaMinus+1/LambdaPlus]/2
          Diff(InvLambdaMinus,InvLambdaPlus,Delto);
          Mult(SHalf,Delto);                        --- Delto = [1/LambdaMinus-1/LambdaPlus]/2

          Prod(Delto,CN,Tmp1);
          Prod(Delto,SN,Tmp2);
          Mult(Sigmo,CN);
          Mult(Sigmo,SN);
          if P=0 then
            Add(Tmp2,CN);     --- CN := Sigmo*CN_old + Delto*SN_old
            Add(Tmp1,SN);     --- SN := Sigmo*SN_old + Delto*CN_old
          else
            Sub(Tmp2,CN);     --- CN := Sigmo*CN_old - Delto*SN_old
            Sub(Tmp1,SN);     --- SN := Sigmo*SN_old - Delto*CN_old
          end if;
        end if;
      end;
    end loop;
    if Not_Trunc then
      declare
        CE: Errs1 renames C.E;
        SE: Errs1 renames S.E;
        CEN,SEN: Radius;
      begin
        for N in EFreq loop
          CEN := CE(N);
          SEN := SE(N);
          CE(N) := QSig(N)*CE(N)+QDel(N)*SEN;
          SE(N) := QSig(N)*SE(N)+QDel(N)*CEN;
        end loop;
      end;
    end if;
  end QP_InvHelm;

  procedure InvQP(F: in out Spheric; QSigDel: in CCErrs; QLam: in EArray) is
    FC: CArray renames F.C;
    K: Integer;
  begin
    if not Mixed then raise Undefined; end if;
    ProjAll(F);
    for L in 0 .. Size loop
      if not IsZero(FC(L,L)) then raise Not_Implemented; end if;
      for M in 1 .. L loop
        K := L-M;
        declare
          QSig: Errs1 renames QSigDel(L,K);
          QDel: Errs1 renames QSigDel(K,L);
        begin
          QP_InvHelm(L,M,Par,QSig,QDel,FC(L,K),FC(K,L));
        end;
      end loop;
    end loop;
    if Not_Trunc then
      if Symm>=0 then raise Not_Implemented; end if;
      declare
        FE: EArray renames F.E;
      begin
        for L in 1 .. SizeE loop FE(L) := FE(L)/QLam(L); end loop;
      end;
    end if;
  end InvQP;

  procedure InvQP(F: in out Spheric) is
  begin
    InvQP(F,QP_SigmaDelta,QP_Lambda); --- global vars QP_SigDel and QP_Lambda become const
  end InvQP;

  ------------------
  --- symmetries ---
  ------------------

  procedure FlipNorthSouth(F: in out Spheric) is
    --- factor (-1)^{l-m}
    C: CArray renames F.C;
    K: Integer := 1;
  begin
    while K<=Size loop          -- K=L-M
      Neg(C(K,K));              -- M=0
      for L in K+1 .. Size loop -- M>0
        Neg(C(L,K));
        Neg(C(K,L));
      end loop;
      K := K+2;
    end loop;
  end FlipNorthSouth;

  procedure Reflect0(F: in out Spheric) is
    --- x --> -x, factor (-1)^l
    C: CArray renames F.C;
    L: Integer := 1;
  begin
    while L<=Size loop
      Neg(C(L,L));           -- M=0
      for K in 0 .. L-1 loop -- K=L-M
        Neg(C(L,K));
        Neg(C(K,L));
      end loop;
      L := L+2;
    end loop;
  end Reflect0;

  procedure PhiRotatePi(F: in out Spheric) is
    --- factor (-1)^m
    C: CArray renames F.C;
    M: Integer := 1;
  begin
    while M<=Size loop
      for L in M .. Size loop
        Neg(C(L,L-M));
        Neg(C(L-M,L));
      end loop;
      M := M+2;
    end loop;
  end PhiRotatePi;

  procedure ProjReflect0(Ref: in MinusOneZeroOne; F: in out Spheric) is
    --- Ref in {-1,0,1} --> {odd,neither,even} restriction on L
  begin
    if Ref/=0 then
      declare
        C: CArray renames F.C;
        L: Natural;
      begin
        if Ref>0 then L := 1; else L := 0; end if;
        while L<=Size loop
          FillZero(C(L,L));
          for K in 0 .. L-1 loop
            FillZero(C(L,K));
            FillZero(C(K,L));
          end loop;
          L := L+2;
        end loop;
      end;
      if Not_Trunc then
        declare
          E: EArray renames F.E;
          L: Natural;
        begin
          if Ref>0 then L := 1; else L := 0; end if;
          while L<SizeE loop
            E(L+1) := E(L+1)+E(L);
            E(L) := Zero;
            L := L+2;
          end loop;
        end;
      end if;
    end if;
  end ProjReflect0;

  procedure ProjPhi(Ref: in MinusOneZeroOne; F: in out Spheric) is
    --- Ref in {-1,0,1} --> {odd,neither,even} in Phi
  begin
    if Ref>0 then             --- even
      declare
        C: CArray renames F.C;
      begin
        for L in 1 .. Size loop
          for K in 0 .. L-1 loop FillZero(C(K,L)); end loop; -- K=L-M
        end loop;
      end;
    elsif Ref<0 then          --- odd
      declare
        C: CArray renames F.C;
      begin
        for L in 0 .. Size loop
          for K in 0 .. L loop FillZero(C(L,K)); end loop;   -- K=L-M
        end loop;
      end;
      if Not_Trunc then
        declare
          E: EArray renames F.E;
        begin
          E(1) := E(1)+E(0);
          E(0) := Zero;
        end;
      end if;
    end if;
  end ProjPhi;

  procedure ProjNorthSouth(Ref: in MinusOneZeroOne; F: in out Spheric) is
    --- Ref>0: SetZero C(L,K) and C(K,L) if K is odd
    --- Ref<0: SetZero C(L,K) and C(K,L) if K is even
  begin
    if Ref/=0 then
      declare
        C: CArray renames F.C;
        K: Integer;
      begin
        if Ref>0 then K := 1; else K := 0; end if;
        while K<=Size loop          -- K=L-M
          FillZero(C(K,K));         -- M=0
          for L in K+1 .. Size loop -- M>0
            FillZero(C(L,K));
            FillZero(C(K,L));
          end loop;
          K := K+2;
        end loop;
      end;
      if Not_Trunc and then Ref<0 then -- no constant term
        declare
          E: EArray renames F.E;
        begin
          E(1) := E(1)+E(0);
          E(0) := Zero;
        end;
      end if;
    end if;
  end ProjNorthSouth;

  procedure ProjMod(Base: in Positive; Dig: in Natural; F: in out Spheric) is
    --- keep only terms with (M mod Base)=Dig
    C: CArray renames F.C;
  begin
    for M in CRange loop
      if (M mod Base)/=Dig then
        for L in M .. Size loop
          FillZero(C(L,L-M));
          FillZero(C(L-M,L));
        end loop;
      end if;
    end loop;
    if Not_Trunc and then Dig>0 then --- no constant term
      declare
        E: EArray renames F.E;
      begin
        E(1) := E(1)+E(0);
        E(0) := Zero;
      end;
    end if;
  end ProjMod;

  procedure ProjM0(F: in out Spheric) is
    --- invariant under Phi rotation, M=0
    C: CArray renames F.C;
  begin
    for L in 1 .. Size loop
      for K in 0 .. L-1 loop -- M=L-K
        FillZero(C(L,K));
        FillZero(C(K,L));
      end loop;
    end loop;
  end ProjM0;

  procedure ProjSymm(Rot: in Integer; F: in out Spheric) is
  begin
    if Rot>1 then
      ProjMod(Rot,0,F);
    elsif Rot<0 then
      ProjMod(2*abs(Rot),abs(Rot),F);
    elsif Rot=0 then
       ProjM0(F);
    end if;
  end ProjSymm;

  procedure ProjFreq(P: in MinusOneZeroOne; F: in out Spheric) is
    --- time frequencies N
    C: CArray renames F.C;
  begin
    if P<0 then         --- odd N
      if FDeg=0 then
        FillZero(F);
      else
        for I in CRange loop
          for J in CRange loop ProjOddFreq(C(I,J)); end loop;
        end loop;
      end if;
    elsif P>0 then       --- even N
      for I in CRange loop
        for J in CRange loop ProjEvenFreq(C(I,J)); end loop;
      end loop;
    end if;
  end ProjFreq;

  procedure ProjParity(P: in Parity; Mix: in Boolean; F: in out Spheric; Check: in Boolean := True) is
    --- Mix=False: parity P under  t     --> -t
    --- Mix=True:  parity P under (t,phi)-->(-t,-phi)
    PC: Parity renames P;                     --- desired time-parity of cos(m*phi) term
    PS: constant Parity := Choose(Mix,1-P,P); --- desired time-parity of sin(m*phi) term
    C: CArray renames F.C;
  begin
    for L in CRange loop
      ProjPar(PC,C(L,L),Check);   --- cos(0*phi){cos,sin}(n*t)
      for K in 0 .. L-1 loop      --- M=L-K
        ProjPar(PC,C(L,K),Check); --- cos(m*phi){cos,sin}(n*t)
        ProjPar(PS,C(K,L),Check); --- sin(m*phi){cos,sin}(n*t)
      end loop;
    end loop;
  end ProjParity;

  procedure ProjStat(F: in out Spheric) is
    --- time-stationary functions
    C: CArray renames F.C;
  begin
    if FDeg>0 then
      for I in CRange loop
        for J in CRange loop ProjConst(C(I,J)); end loop;
      end loop;
    end if;
  end ProjStat;

  procedure ProjAll(F: in out Spheric) is
    Check: constant Boolean := True;
  begin
    ProjFreq(FreqPar,F);
    if Full then
      null;               -- caution: could be inconsistent with RefPhi, RefTh, Symm
    else
      ProjPhi(RefPhi,F);
      ProjTheta(RefTh,F);
      ProjSymm(Symm,F);
    end if;
    ProjParity(Par,Mixed,F,Check); -- raise exceeption if inconsistent
  end ProjAll;

  procedure Parities(F: in Spheric; CosEven,CosOdd,SinEven,SinOdd: out Boolean) is
    --- CosEven is True iff t-parity of all cos(m*phi) terms is even
    --- CosOdd  is True iff t-parity of all cos(m*phi) terms is odd
    --- SinEven is True iff t-parity of all sin(m*phi) terms is even
    --- SinOdd  is True iff t-parity of all sin(m*phi) terms is odd
    --- notice: a function that is both even and odd is zero
    C: CArray renames F.C;
    Even,Odd,Foo: Boolean;
  begin
    CosEven := True;
    SinEven := True;
    SinOdd := True;
    if FDeg=0 then
      CosOdd := IsZero(C);
    else
      CosOdd := True;
      for L in CRange loop
        for K in 0 .. L loop           -- M=L=K
          Parities(C(L,K),Even,Odd);   -- (L,K) cos(m*phi)
          CosEven := CosEven and Even;
          CosOdd := CosOdd and Odd;
        end loop;
        for K in 0 .. L-1 loop         -- M=L=K
          Parities(C(K,L),Even,Odd);   -- (K,L) sin(m*phi)
          SinEven := SinEven and Even;
          SinOdd := SinOdd and Odd;
        end loop;
      end loop;
      Foo := (not (CosEven or CosOdd)) or (not (SinEven or SinOdd));
      if Foo then
        Show_Parities("",F,False);
        raise Parity_Error;            -- neither even nor odd
      end if;
    end if;
  end Parities;

  procedure CheckParities(F: in Spheric; Mix: in Logical := Uncertain) is
    CosEven,CosOdd,SinEven,SinOdd: Boolean;
  begin
    Parities(F,CosEven,CosOdd,SinEven,SinOdd);
    if Mix=Uncertain then           -- no more checks
      null;
    elsif (CosEven and CosOdd) then -- all cos(m*phi) terms are zero
      null;
    elsif (SinEven and SinOdd) then -- all sin(m*phi) terms are zero
      null;
    elsif Mix=False then
      if CosEven/=SinEven then
        Show_Parities("",F,False);
        raise Parity_Error;
      end if;
    else                            -- Mix=True, swap SinEven and SinOdd
      if CosEven/=SinOdd then
        Show_Parities("",F,False);
        raise Parity_Error;
      end if;
    end if;
  end CheckParities;

  procedure Sanitize_Parities(F: in out Spheric) is
    --- independent of the Mixed value
    C: CArray renames F.C;
    CosEven,CosOdd,SinEven,SinOdd: Boolean;
    P: Parity;
  begin
    Parities(F,CosEven,CosOdd,SinEven,SinOdd);
    if CosEven/=CosOdd then
      if CosEven then P := 0; else P := 1; end if;
      for L in 0 .. Size loop
        for K in 0 .. L loop C(L,K).P := P; end loop;
      end loop;
    end if;
    if SinEven/=SinOdd then
      if SinEven then P := 0; else P := 1; end if;
      for L in 1 .. Size loop
        for K in 0 .. L-1 loop C(K,L).P := P; end loop;
      end loop;
    end if;
  end Sanitize_Parities;

  procedure Guess_Parity(F: in Spheric; Mix: in Boolean; P: out Parity; Is_Zero: out Boolean) is
    --- correct if C is consistent, but C=0 is ambiguous
    CP_Even,SP_Even,Odd, CZero,SZero: Boolean;
    C: CArray renames F.C;
  begin
Find_CP:
    for L in 0 .. Size loop
      for K in 0 .. L loop              -- M=L-K
        Parities(C(L,K),CP_Even,Odd);   -- cos(m*phi)
        exit Find_CP when CP_Even/=Odd; -- first nonzero C(L,K)
      end loop;
    end loop Find_CP;                   -- CP_Even=Odd=False cannot occcur
    CZero := CP_Even and Odd;           -- all cos(m*phi) terms zero
                                        ---
Find_SP:
    for L in 1 .. Size loop
      for K in 0 .. L-1 loop            -- M=L-K
        Parities(C(K,L),SP_Even,Odd);   -- sin(m*phi)
        exit Find_SP when SP_Even/=Odd; -- first nonzero C(K,L)
      end loop;
    end loop Find_SP;
    SZero := SP_Even and Odd;           -- all sin(m*phi) terms zero
                                        ---
    if Mix then SP_Even := not SP_Even; end if;    -- mixed parities now
    if CZero then                                  -- no cos(m*phi) terms
      if SP_Even then P := 0; else P := 1; end if;
      Is_Zero := SZero;
    elsif SZero then                               -- no sin(m*phi) terms
      if CP_Even then P := 0; else P := 1; end if;
      Is_Zero := False;
    elsif SP_Even=CP_Even then                     -- same (mixed) parity
      if CP_Even then P := 0; else P := 1; end if;
      Is_Zero := False;
    else
      Show_Parities("",F,False);
      raise Parity_Error;
    end if;
  end Guess_Parity;

  procedure Simple_Random(F: in out Spheric; P: in Parity; Mix: in Boolean; Fac: in Radius := Two) is
    --- if not Mix: P is the parity for t --> -t
    --- if Mix:     P is the parity for (t,Phi) --> (-t,-phi)
    PC: Parity renames P;                     --- time-parity of cos(m*phi) term
    PS: constant Parity := Choose(Mix,1-P,P); --- time-parity of sin(m*phi) term
    NewRad: constant Radius := Fac*Rho1;
    NewRho: constant Radius := Fac*Rho2;
    C: CArray renames F.C;
    Prod: Radius := One;
  begin
    for L in CRange loop
      for K in 0 .. L-1 loop             -- K=L-M
        Simple_Random(NewRad,PC,C(L,K)); -- cos(m*phi){sin,cos}(n*t)
        Mult(Prod,C(L,K));
        Simple_Random(NewRad,PS,C(K,L)); -- sin(m*phi){sin,cos}(n*t)
        Mult(Prod,C(K,L));
      end loop;
      Simple_Random(NewRad,PC,C(L,L));   -- cos(0*phi){sin,cos}(n*t)
      Mult(Prod,C(L,L));
      Prod := Prod/NewRho;
    end loop;
    if Not_Trunc then F.E := (others => Zero); end if;
  end Simple_Random;

  procedure NumTrunc(SizC,SizS: in Natural; F: in out Spheric) is
    --- rarely used, so...
    C: CArray renames F.C;
  begin
    for I in CRange loop
      for J in CRange loop
        if I>SizS or else J>SizS then
          FillZero(C(I,J));
        else
          NumTrunc(SizC,C(I,J));
        end if;
      end loop;
    end loop;
    if Not_Trunc then F.E := (others => Zero); end if;
  end NumTrunc;

  procedure NumTrunc(SizC,SizS: in Natural; F1: in Spheric; F2: in out Spheric) is
  begin
    Copy(F1,F2);
    NumTrunc(SizC,SizS,F2);
  end NumTrunc;

  ----------------
  --- plotting ---
  ----------------

  procedure Make_PlotData(F: in Spheric; P: in out CMatrix2) is
    --- Theta in [0,Pi], Phi in [-Pi,Pi]
    Plast1: constant Positive := P'Last(1);     --- assuming P'First(1)=0
    Plast2: constant Positive := P'Last(2);     --- assuming P'First(2)=0
    DTheta: constant Flt := PiFull/Flt(PLast1); --- first coordinate is Theta
    DPhi:   constant Flt := TwoPi/Flt(PLast2);  --- second coordinate is Phi
    Theta,Phi: Flt;
    STheta,SPhi: Scalar;
  begin
    for N1 in 0 .. PLast1 loop
      Theta := Flt(N1)*DTheta;
      Assign(Theta,STheta);
      for N2 in 0 .. PLast2 loop
        Phi := Flt(N2)*DPhi-PiFull;
        Assign(Phi,SPhi);
        Val2(F,STheta,SPhi,P(N1,N2));
      end loop;
    end loop;
  end Make_PlotData;

  procedure SinProjection(Theta,Phi: in Flt; X,Y: out Flt) is
    --- Theta in [0,Pi], Phi in [-Pi,Pi]
    use Flt_EF;
  begin
    X := Phi*Sin(Theta);
    Y := PiHalf-Theta;
  end SinProjection;

  procedure Plot(FileName: in String; P: in CMatrix2; T: in Flt) is
    --- Theta in [0,Pi], Phi in [-Pi,Pi]
    Digs: constant Positive := 5;
    Plast1: constant Positive := P'Last(1);     --- assuming P'First(1)=0
    Plast2: constant Positive := P'Last(2);     --- assuming P'First(2)=0
    DTheta: constant Flt := PiFull/Flt(PLast1); --- first coordinate is Theta
    DPhi:   constant Flt := TwoPi/Flt(PLast2);  --- second coordinate is Phi
    Theta,Phi,X,Y,Z: Flt;
    F: File_Type;
    package Txt_IO renames Ada.Text_IO;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    for N1 in 0 .. PLast1 loop
      Theta := Flt(N1)*DTheta;
      for N2 in 0 .. PLast2 loop
        Phi := Flt(N2)*DPhi-PiFull;
        SinProjection(Theta,Phi,X,Y);
        Z := NumVal(P(N1,N2),T);
        ---
        Flt_IO.Put(F,X,3,Digs,3);
        Txt_IO.Put(F," ");
        Flt_IO.Put(F,Y,3,Digs,3);
        Txt_IO.Put(F," ");
        Flt_IO.Put(F,Z,3,Digs,3);
        if Plot_Newline then New_Line(F); end if;
      end loop;
      if Plot_Newline then New_Line(F); end if;
    end loop;
    Close(F);
  end Plot;

  procedure Plot(FileName: in String; F: in Spheric; T: in Flt; ThetaSteps: in Positive := 128; PhiSteps: in Positive := 64) is
    --- do not use in loops over T
    P: CMatrix2(0 .. ThetaSteps, 0 .. PhiSteps);
  begin
    Make_PlotData(F,P);
    Plot(FileName,P,T);
  end Plot;

  procedure Plot(Stem: in String; F: in Spheric; ThetaSteps,PhiSteps,TSteps: in Positive) is
    --- writes TSteps files
    --- if TSteps is 64, then the files are Stem000, Stem001, ..., Stem063
    DT: constant Flt := TwoPi/Flt(TSteps);
    T: Flt;
    P: CMatrix2(0 .. ThetaSteps, 0 .. PhiSteps);
  begin
    Make_PlotData(F,P);
    for N in 0 .. TSteps-1 loop
      T := Flt(N)*DT;
      if Plot_Newline then
        Plot(Stem & Strng0(N,3),P,T);
      else
        Plot(Stem & Strng(N),P,T);
      end if;
      exit when FDeg=0;
    end loop;
  end Plot;

  -------------
  --- modes ---
  -------------

  procedure Norm(A: in TSMode; S: in out Scalar) is
  begin
    if A.CErr or else A.SErr then
      Copy(SOne,S);  --- error modes have norm 1
    else
      declare
        Fac,Tmp: Scalar;
      begin
        IPower(A.N,SRad,Fac,Tmp);
        IPower(IMax(A.I,A.J),SRho,S,Tmp); -- IMax(A.I,A.J)=L
        Mult(Fac,S);
      end;
    end if;
  end Norm;

  procedure MaxNorm(A: in TSMode; S: out Radius) is
  begin
    if A.CErr or else A.SErr then
      S := One;      --- error modes have norm 1
    else
      S := RadPower(A.N)*Rho2Power(IMax(A.I,A.J)); -- IMax(A.I,A.J)=L
    end if;
  end MaxNorm;

  procedure Assign(A: in TSMode; F: in out Spheric) is
    N: Natural renames A.N;
    I: ERange renames A.I;
    J: ERange renames A.J;
  begin
    SetZero(F);
    if A.CErr then
      F.C(I,J).E(N) := InvRho2Power(IMax(I,J));
      if A.SErr then raise Inconsistent_Modes; end if;
    elsif A.SErr then
      F.E(IMax(I,J)) := One;
    else
      Copy(SOne,F.C(I,J).C(N));
    end if;
  end Assign;

  procedure AddProd(A: in TSMode; S: in Scalar; F: in out Spheric; Check: in Boolean) is
    pragma Unreferenced(Check);
  begin
    if not IsZero(S) then
      if A.CErr or else A.SErr then raise Not_Implemented; end if;
      Add(S,F.C(A.I,A.J).C(A.N));
    end if;
  end AddProd;

  procedure Extract_Coeff(A: in TSMode; F: in out Spheric; S: in out Scalar) is
    --- assuming A.CErr=False and A.SErr=False
    --- S encloses the coefficient
  begin
    SetZero(S);
    Swap(F.C(A.I,A.J).C(A.N),S);
  end Extract_Coeff;

  procedure Extract_Err(A: in TSMode; F: in out Spheric; S: in out Scalar) is
    pragma Unreferenced(A);
    pragma Unreferenced(F);
    pragma Unreferenced(S);
  begin
    raise Not_Implemented;
  end Extract_Err;

  procedure Extract(A: in TSMode; F: in out Spheric; S: in out Scalar; Check: in Boolean) is
    --- assuming A is consistent
    pragma Unreferenced(Check);
  begin
    if A.CErr or else A.SErr then
      Extract_Err(A,F,S);
    else
      Extract_Coeff(A,F,S);
    end if;
  end Extract;

  procedure Show(A: in TSMode; NewLine: in Boolean := False) is
  begin
    Show1("   N = ",A.N);
    Show1("   I = ",A.I);
    Show1("   J = ",A.J);
    Show1("CErr = ",A.CErr);
    Show1("SErr = ",A.SErr);
    if NewLine then New_Line; end if;
  end Show;

  function Symm_Check(A: TSMode) return Boolean is
    N: Natural renames A.N;  -- frequency
    I: ERange renames A.I;
    J: ERange renames A.J;
    OddPhi: constant Boolean := I<J;
    EvenPhi: constant Boolean := not OddPhi;
    L: constant Natural := IMax(I,J);
    K: constant Natural := IMin(I,J); --- K=L-M
    M: constant Natural := L-K;
  begin
    if FDeg=0 and then N>0 then return False; end if;                 -- time-indepedent solutions
    if N=0 then                                                       -- avoid sin(0)
      if Mixed then
        if OddPhi and then Par=0 then return False; end if;
        if EvenPhi and then Par=1 then return False; end if;
      else
        if Par=1 then return False; end if;
      end if;
    end if;
    if (FreqPar= 1) and then (N mod 2)= 1 then return False; end if;  -- only even frequencies
    if (FreqPar=-1) and then (N mod 2)= 0 then return False; end if;  -- only odd frequencies
    if L=0 then return False; end if;                                 -- constant mode in null space
    if A.SErr then return True; end if;                               -- Spheric errors have no symmetries
    if Full then return True; end if;                                 -- no symmetries
    if RefPhi= 1 and then OddPhi then return False; end if;           -- want PhiEven
    if RefPhi=-1 and then EvenPhi then return False; end if;          -- want PhiOdd
    if RefTh = 1 and then (K mod 2)=1 then return False; end if;      -- want North/South even
    if RefTh =-1 and then (K mod 2)=0 then return False; end if;      -- want North/South odd
    if Symm=1 then return True; end if;                               -- no rotation symmetry
    if Symm=0 then return M=0; end if;                                -- invariant wrt phi rotation
    if Symm>1 then return (M mod Symm)=0; end if;                     -- Symm>1 rotation symm
    if Symm<0 then return (M mod (2*abs(Symm)))=abs(Symm); end if;    -- Symm<0 rotation antisymm
    return True;
  end Symm_Check;

  function Make(CSize: Natural; SSize: Positive; No_Errs: Boolean) return TSModes is
    --- C* refers to CosSins1, S* refers to Spheric
    Exclude_Errs: constant Boolean := No_Errs or Trunc;
    Include_Errs: constant Boolean := not Exclude_Errs;
    CSise: constant Natural := 2*(CSize/2); --- even (necessary?)
    SSise: constant Positive := 2*(SSize/2); --- even (necessary?)
    CSise1: constant Positive := CSise+1;
    SSise1: constant Positive := SSise+1;

    procedure Build(M: out TSModes; Count: out Natural) is

      procedure Set_Mode(N,I,J: in Natural; CErr,SErr: in Boolean) is
        Tmp: constant TSMode := (N,I,J,CErr,SErr);
      begin
        if Symm_Check(Tmp) then
          Count := Count+1;
          if Count <= M'Last then M(Count) := Tmp; end if;
        end if;
      end Set_Mode;

    begin
      Count := 0;
      for I in 0 .. SSise loop
        for J in 0 .. SSise loop
           for N in 0 .. CSise loop
              --if Exclude_Errs then
                 Set_Mode(N,I,J,False,False);        --- coefficients
              --end if;
          end loop;
        end loop;
      end loop;
      if Include_Errs then
        if FDeg>0 then                         --- no CErr when FDeg=0
          for I in 0 .. SSise loop
            for J in 0 .. SSise loop
              Set_Mode(CSise1,I,J,True,False); --- CErr (CosSin1)
            end loop;
          end loop;
        end if;
        Set_Mode(0,SSise1,SSise1,False,True);  --- SErr (Spheric)
      end if;
    end Build;

    Count: Natural;
  begin
    declare
      M: TSModes(1..1);
    begin
      Build(M,Count);
    end;
    declare
      M: TSModes(1..Count);
    begin
      Build(M,Count);
      return M;
    end;
  end Make;

begin

  if Map=2 or Map=3 then
    if Trunc then
      Wave_Delta := (others => Zero);
    else
      Make_Wave_Del(Wave_Delta);
    end if;
  elsif Map=4 then
    if Trunc then
      QP_Lambda := (others => Zero);
      QP_SigmaDelta := (others => (others => (others => Zero)));
    else
      for L in 1 .. Size loop
        for M in 1 .. L loop
          declare
            QSig: Errs1 renames QP_SigmaDelta(L,L-M);
            QDel: Errs1 renames QP_SigmaDelta(L-M,L);
          begin
            QP_SigDel(SBeta,L,M,QSig,QDel);
          end;
        end loop;
      end loop;
      QP_Lam(BetaI,BetaJ,abs(Symm),QP_Lambda);
    end if;
  end if;

end TSpherics;
